# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Weisfeiler-Lehman Network (WLN) for ranking candidate products
# pylint: disable= no-member, arguments-differ, invalid-name

import torch
import torch.nn as nn

from dgl.nn.pytorch import SumPooling

from ..gnn.wln import WLN

__all__ = ['WLNReactionRanking']

# pylint: disable=W0221, E1101
class WLNReactionRanking(nn.Module):
    r"""Weisfeiler-Lehman Network (WLN) for Candidate Product Ranking

    The model is introduced in `Predicting Organic Reaction Outcomes with
    Weisfeiler-Lehman Network <https://arxiv.org/abs/1709.04555>`__ and then
    further improved in `A graph-convolutional neural network model for the
    prediction of chemical reactivity
    <https://pubs.rsc.org/en/content/articlelanding/2019/sc/c8sc04228d#!divAbstract>`__

    The model updates representations of nodes in candidate products with WLN and predicts
    the score for candidate products to be the real product.

    Parameters
    ----------
    node_in_feats : int
        Size for the input node features.
    edge_in_feats : int
        Size for the input edge features.
    node_hidden_feats : int
        Size for the hidden node representations. Default to 500.
    num_encode_gnn_layers : int
        Number of WLN layers for updating node representations.
    """
    def __init__(self,
                 node_in_feats,
                 edge_in_feats,
                 node_hidden_feats=500,
                 num_encode_gnn_layers=3):
        super(WLNReactionRanking, self).__init__()

        self.gnn = WLN(node_in_feats=node_in_feats,
                       edge_in_feats=edge_in_feats,
                       node_out_feats=node_hidden_feats,
                       n_layers=num_encode_gnn_layers,
                       set_comparison=False)
        self.diff_gnn = WLN(node_in_feats=node_hidden_feats,
                            edge_in_feats=edge_in_feats,
                            node_out_feats=node_hidden_feats,
                            n_layers=1,
                            project_in_feats=False,
                            set_comparison=False)
        self.readout = SumPooling()
        self.predict = nn.Sequential(
            nn.Linear(node_hidden_feats, node_hidden_feats),
            nn.ReLU(),
            nn.Linear(node_hidden_feats, 1)
        )

    def forward(self, reactant_graph, reactant_node_feats, reactant_edge_feats,
                product_graphs, product_node_feats, product_edge_feats,
                candidate_scores, batch_num_candidate_products):
        r"""Predicts the score for candidate products to be the true product

        Parameters
        ----------
        reactant_graph : DGLGraph
            DGLGraph for a batch of reactants.
        reactant_node_feats : float32 tensor of shape (V1, node_in_feats)
            Input node features for the reactants. V1 for the number of nodes.
        reactant_edge_feats : float32 tensor of shape (E1, edge_in_feats)
            Input edge features for the reactants. E1 for the number of edges in
            reactant_graph.
        product_graphs : DGLGraph
            DGLGraph for the candidate products in a batch of reactions.
        product_node_feats : float32 tensor of shape (V2, node_in_feats)
            Input node features for the candidate products. V2 for the number of nodes.
        product_edge_feats : float32 tensor of shape (E2, edge_in_feats)
            Input edge features for the candidate products. E2 for the number of edges
            in the graphs for candidate products.
        candidate_scores : float32 tensor of shape (B, 1)
            Scores for candidate products based on the model for reaction center prediction
        batch_num_candidate_products : list of int
            Number of candidate products for the reactions in the batch

        Returns
        -------
        float32 tensor of shape (B, 1)
            Predicted scores for candidate products
        """
        # Update representations for nodes in both reactants and candidate products
        batch_reactant_node_feats = self.gnn(
            reactant_graph, reactant_node_feats, reactant_edge_feats)
        batch_product_node_feats = self.gnn(
            product_graphs, product_node_feats, product_edge_feats)

        # Iterate over the reactions in the batch
        reactant_node_start = 0
        product_graph_start = 0
        product_node_start = 0
        batch_diff_node_feats = []

        for i, num_candidate_products in enumerate(batch_num_candidate_products):
            reactant_node_end = reactant_node_start + reactant_graph.batch_num_nodes[i]
            product_graph_end = product_graph_start + num_candidate_products
            product_node_end = product_node_start + sum(
                product_graphs.batch_num_nodes[product_graph_start: product_graph_end])

            # (N, node_out_feats)
            reactant_node_feats = batch_reactant_node_feats[reactant_node_start:
                                                            reactant_node_end, :]
            product_node_feats = batch_product_node_feats[product_node_start: product_node_end, :]

            old_feats_shape = reactant_node_feats.shape
            # (1, N, node_out_feats)
            expanded_reactant_node_feats = reactant_node_feats.reshape((1,) + old_feats_shape)
            # (B, N, node_out_feats)
            expanded_reactant_node_feats = expanded_reactant_node_feats.expand(
                (num_candidate_products,) + old_feats_shape)
            # (B, N, node_out_feats)
            candidate_product_node_feats = product_node_feats.reshape(
                (num_candidate_products,) + old_feats_shape)

            # Get the node representation difference between candidate products and reactants
            diff_node_feats = candidate_product_node_feats - expanded_reactant_node_feats
            diff_node_feats = diff_node_feats.reshape(-1, diff_node_feats.shape[-1])
            batch_diff_node_feats.append(diff_node_feats)

            reactant_node_start = reactant_node_end
            product_graph_start = product_graph_end
            product_node_start = product_node_end

        batch_diff_node_feats = torch.cat(batch_diff_node_feats, dim=0)
        # One more GNN layer for message passing with the node representation difference
        diff_node_feats = self.diff_gnn(product_graphs, batch_diff_node_feats, product_edge_feats)
        candidate_product_feats = self.readout(product_graphs, diff_node_feats)

        return self.predict(candidate_product_feats) + candidate_scores



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# GraphSAGE
# pylint: disable= no-member, arguments-differ, invalid-name

import torch.nn as nn
import torch.nn.functional as F

from dgl.nn.pytorch import SAGEConv

__all__ = ['GraphSAGE']

# pylint: disable=W0221, C0103
class GraphSAGE(nn.Module):
    r"""GraphSAGE from `Inductive Representation Learning on Large Graphs
    <https://arxiv.org/abs/1706.02216>`__

    Parameters
    ----------
    in_feats : int
        Number of input node features.
    hidden_feats : list of int
        ``hidden_feats[i]`` gives the size of node representations after the i-th GraphSAGE layer.
        ``len(hidden_feats)`` equals the number of GraphSAGE layers.  By default, we use
        ``[64, 64]``.
    activation : list of activation functions or None
        If not None, ``activation[i]`` gives the activation function to be used for
        the i-th GraphSAGE layer. ``len(activation)`` equals the number of GraphSAGE layers.
        By default, ReLU is applied for all GraphSAGE layers.
    dropout : list of float or None
        ``dropout[i]`` decides the dropout probability on the output of the i-th GraphSAGE layer.
        ``len(dropout)`` equals the number of GraphSAGE layers. By default, no dropout is
        performed for all layers.
    aggregator_type : list of str
        ``aggregator_type[i]`` decides the aggregator type for the i-th GraphSAGE layer, which
        can be one of ``'mean'``, ``'gcn'``, ``'pool'``, ``'lstm'``. By default, we use
        ``'mean'`` for all layers.
    """
    def __init__(self,
                 in_feats,
                 hidden_feats=None,
                 activation=None,
                 dropout=None,
                 aggregator_type=None):
        super(GraphSAGE, self).__init__()

        if hidden_feats is None:
            hidden_feats = [64, 64]

        n_layers = len(hidden_feats)
        if activation is None:
            activation = [F.relu for _ in range(n_layers)]
        if dropout is None:
            dropout = [0. for _ in range(n_layers)]
        if aggregator_type is None:
            aggregator_type = ['mean' for _ in range(n_layers)]
        lengths = [len(hidden_feats), len(activation), len(dropout), len(aggregator_type)]
        assert len(set(lengths)) == 1, 'Expect the lengths of hidden_feats, activation, ' \
                                       'dropout and aggregator_type to be the same, ' \
                                       'got {}'.format(lengths)

        self.hidden_feats = hidden_feats
        self.gnn_layers = nn.ModuleList()
        for i in range(n_layers):
            self.gnn_layers.append(SAGEConv(in_feats, hidden_feats[i], aggregator_type[i],
                                            dropout[i], activation[i]))
            in_feats = hidden_feats[i]

    def reset_parameters(self):
        """Reinitialize model parameters."""
        for gnn in self.gnn_layers:
            gnn.reset_parameters()

    def forward(self, g, feats):
        """Update node representations.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs
        feats : FloatTensor of shape (N, M1)
            * N is the total number of nodes in the batch of graphs
            * M1 is the input node feature size, which equals in_feats in initialization

        Returns
        -------
        feats : FloatTensor of shape (N, M2)
            * N is the total number of nodes in the batch of graphs
            * M2 is the output node representation size, which equals
              hidden_sizes[-1] in initialization.
        """
        for gnn in self.gnn_layers:
            feats = gnn(g, feats)
        return feats



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Utils for analyzing a collection of molecules

from collections import defaultdict
from multiprocessing import Pool

import itertools
import numpy as np

from rdkit import Chem

__all__ = ['analyze_mols']

def summarize_a_mol(mol):
    """Summarize a molecule

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        Input molecule

    Returns
    -------
    dict
        Summarized info for a molecule
    """
    summary = {
        'num_atoms': mol.GetNumAtoms(),
        'num_bonds': mol.GetNumBonds(),
        'num_rings': len(Chem.GetSymmSSSR(mol)),
        'atom_type': set(),
        'degree': set(),
        'total_degree': set(),
        'explicit_valence': set(),
        'implicit_valence': set(),
        'hybridization': set(),
        'total_num_h': set(),
        'formal_charge': set(),
        'num_radical_electrons': set(),
        'aromatic_atom': set(),
        'chirality_tag': set(),
        'bond_type': set(),
        'conjugated_bond': set(),
        'bond_stereo_configuration': set(),
        'bond_direction': set()
    }

    for atom in mol.GetAtoms():
        summary['atom_type'].add(atom.GetSymbol())
        summary['degree'].add(atom.GetDegree())
        summary['total_degree'].add(atom.GetTotalDegree())
        summary['explicit_valence'].add(atom.GetExplicitValence())
        summary['implicit_valence'].add(atom.GetImplicitValence())
        summary['hybridization'].add(str(atom.GetHybridization()))
        summary['total_num_h'].add(atom.GetTotalNumHs())
        summary['formal_charge'].add(atom.GetFormalCharge())
        summary['num_radical_electrons'].add(atom.GetNumRadicalElectrons())
        summary['aromatic_atom'].add(atom.GetIsAromatic())
        summary['chirality_tag'].add(str(atom.GetChiralTag()))

    for bond in mol.GetBonds():
        summary['bond_type'].add(str(bond.GetBondType()))
        summary['conjugated_bond'].add(bond.GetIsConjugated())
        summary['bond_stereo_configuration'].add(str(bond.GetStereo()))
        summary['bond_direction'].add(str(bond.GetBondDir()))

    return summary

def count_frequency(values):
    """Count how many times each value appear

    Parameters
    ----------
    values : list

    Returns
    -------
    dict
        Mapping each unique value to the times it appears
    """
    frequency = defaultdict(int)
    for val in values:
        frequency[val] += 1
    return dict(frequency)

def analyze_mols(smiles=None, mols=None, num_processes=1, path_to_export=None):
    r"""Analyze a collection of molecules

    The analysis will 1) filter out invalid molecules and record the valid ones;
    2) record the number of molecules having each particular descriptor/element
    (e.g. single bond). The descriptors/elements considered include:


    If ``path_to_export`` is not None, we will export the analysis results to
    the following files in ``path_to_export``:

        * **valid_canonical_smiles.txt**: A file of canonical SMILES for valid molecules
        * **summary.txt**: A file of all analysis results,
          see the **Examples** section for more details. For summary, we either compute
          mean/std of values or count the frequency that a value appears in molecules.

    Parameters
    ----------
    smiles : list of str, optional
        SMILES strings for a collection of molecules. Can be omitted if mols is not None.
        (Default: None)
    mols : list of rdkit.Chem.rdchem.Mol objects, optional
        RDKit molecule instances for a collection of molecules.
        Can be omitted if smiles is not None. (Default: None)
    num_processes : int, optional
        Number of processes for data analysis. (Default: 1)
    path_to_export : str, optional
        The directory to export analysis results. If not None, we will export the analysis
        results to local files in the specified directory. (Default: None)

    Returns
    -------
    dict
        Summary of the analysis results. For more details, see the **Examples** section.

    Examples
    --------

    >>> from dgllife.utils import analyze_mols

    >>> smiles = ['CCO', 'CC1([C@@H](N2[C@H](S1)[C@@H](C2=O)NC(=O)CC3=CC=CC=C3)C(=O)O)C', '1']
    >>> # Analyze the results and save the results to the current directory
    >>> results = analyze_mols(smiles, path_to_export='.')
    >>> results
    {'num_atoms': [3, 23],                    # Number of atoms in each molecule
     'num_bonds': [2, 25],                    # Number of bonds in each molecule
     'num_rings': [0, 3],                     # Number of rings in each molecule
     'num_input_mols': 3,                     # Number of input molecules
     'num_valid_mols': 2,                     # Number of valid molecules
     'valid_proportion': 0.6666666666666666,  # Proportion of valid molecules
     'cano_smi': ['CCO',                      # Canonical SMILES for valid molecules
     'CC1(C)S[C@@H]2[C@H](NC(=O)Cc3ccccc3)C(=O)N2[C@H]1C(=O)O'],

     # The following items give the number of times each descriptor value appears in molecules

     'atom_type_frequency': {'O': 2, 'C': 2, 'N': 1, 'S': 1},
     'degree_frequency': {1: 2, 2: 2, 3: 1, 4: 1},
     'total_degree_frequency': {2: 2, 4: 2, 1: 1, 3: 1},
     'explicit_valence_frequency': {1: 2, 2: 2, 3: 1, 4: 1},
     'implicit_valence_frequency': {1: 2, 2: 2, 3: 2, 0: 1},
     'hybridization_frequency': {'SP3': 2, 'SP2': 1},
     'total_num_h_frequency': {1: 2, 2: 2, 3: 2, 0: 1},
     'formal_charge_frequency': {0: 2},
     'num_radical_electrons_frequency': {0: 2},
     'aromatic_atom_frequency': {False: 2, True: 1},
     'chirality_tag_frequency': {'CHI_UNSPECIFIED': 2,
     'CHI_TETRAHEDRAL_CCW': 1,
     'CHI_TETRAHEDRAL_CW': 1},
     'bond_type_frequency': {'SINGLE': 2, 'DOUBLE': 1, 'AROMATIC': 1},
     'conjugated_bond_frequency': {False: 2, True: 1},
     'bond_stereo_configuration_frequency': {'STEREONONE': 2},
     'bond_direction_frequency': {'NONE': 2}}
    """
    assert not ((smiles is None) and (mols is None)), \
        'At least one of the smiles and mols should not be None'
    assert num_processes >= 1, \
        'Expect num_processes to be no smaller than 1, got {:d}'.format(num_processes)

    general_items = ['num_atoms', 'num_bonds', 'num_rings']
    atom_items = ['atom_type', 'degree', 'total_degree', 'explicit_valence',
                  'implicit_valence', 'hybridization', 'total_num_h', 'formal_charge',
                  'num_radical_electrons', 'aromatic_atom', 'chirality_tag']
    bond_items = ['bond_type', 'conjugated_bond', 'bond_stereo_configuration', 'bond_direction']

    # Holder of the analysis results
    summary = dict()
    for item in list(itertools.chain.from_iterable([general_items, atom_items, bond_items])):
        summary[item] = []

    # Check validity
    if mols is None:
        mols = []
        for smi in smiles:
            smi_mol = Chem.MolFromSmiles(smi)
            if smi_mol is not None:
                mols.append(smi_mol)
        summary['num_input_mols'] = len(smiles)
        summary['num_valid_mols'] = len(mols)
    else:
        summary['num_input_mols'] = len(mols)
        mols_ = []
        for smi_mol in mols:
            if smi_mol is not None:
                mols_.append(smi_mol)
        mols = mols_
        summary['num_valid_mols'] = len(mols)
    summary['valid_proportion'] = summary['num_valid_mols'] / summary['num_input_mols']

    # Get canonicalized SMILES
    summary['cano_smi'] = [Chem.MolToSmiles(smi_mol) for smi_mol in mols]

    if num_processes == 1:
        summary_per_mol = []
        for smi_mol in mols:
            summary_per_mol.append(summarize_a_mol(smi_mol))
    else:
        with Pool(processes=num_processes) as pool:
            summary_per_mol = pool.map(summarize_a_mol, mols)

    for mol_summary in summary_per_mol:
        for key in general_items:
            summary[key].append(mol_summary[key])
        for key in list(itertools.chain.from_iterable([atom_items, bond_items])):
            summary[key].extend(list(mol_summary[key]))

    for item in list(itertools.chain.from_iterable([atom_items, bond_items])):
        summary[item + '_frequency'] = count_frequency(summary[item])
        del summary[item]

    if path_to_export is None:
        return summary

    # Export the analysis results to local files
    with open(path_to_export + '/valid_canonical_smiles.txt', 'w') as file:
        for smi in summary['cano_smi']:
            file.write(smi + '\n')

    with open(path_to_export + '/summary.txt', 'w') as file:
        file.write('General statistics\n')
        file.write('=' * 60 + '\n')
        file.write('Number of input molecules: {}\n'.format(summary['num_input_mols']))
        file.write('Number of valid molecules: {}\n'.format(summary['num_valid_mols']))
        file.write('Percentage of valid molecules: {} %\n'.format(
            summary['valid_proportion'] * 100))
        file.write('Average number of atoms per molecule: {} +- {}\n'.format(
            np.mean(summary['num_atoms']), np.std(summary['num_atoms'])))
        file.write('Average number of bonds per molecule: {} +- {}\n'.format(
            np.mean(summary['num_bonds']), np.std(summary['num_bonds'])))
        file.write('Average number of rings per molecule: {} +- {}\n'.format(
            np.mean(summary['num_rings']), np.std(summary['num_rings'])))
        file.write('\n')

        file.write('Atom statistics\n')
        file.write('=' * 60 + '\n')
        for item in atom_items:
            file.write('{} frequency: {}\n'.format(item, summary[item + '_frequency']))
        file.write('\n')

        file.write('Bond statistics\n')
        file.write('=' * 60 + '\n')
        for item in bond_items:
            file.write('{} frequency: {}\n'.format(item, summary[item + '_frequency']))
        file.write('\n')

    return summary



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .gnn import *
from .readout import *
from .model_zoo import *
from .pretrain import *



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Lipophilicity from MoleculeNet for the prediction of octanol/water
# distribution coefficient (logD at pH 7.4) of 4200 compounds

import pandas as pd

from dgl.data.utils import get_download_dir, download, _get_dgl_url, extract_archive

from .csv_dataset import MoleculeCSVDataset
from ..utils.mol_to_graph import smiles_to_bigraph

__all__ = ['Lipophilicity']

class Lipophilicity(MoleculeCSVDataset):
    r"""Lipophilicity from MoleculeNet for the prediction of octanol/water
    distribution coefficient (logD at pH 7.4) of 4200 compounds

    This dataset is curated from ChEMBL database containing experimental results
    on octanol/water distribution coefficient (logD at pH=7.4). Due to the importance
    of lipophilicity in membrane permeability and solubility, the task is of high
    importance to drug development.

    References:

        * [1] MoleculeNet: A Benchmark for Molecular Machine Learning.
        * [2] ChEMBL Deposited Data Set - AZ dataset; 2015.

    Parameters
    ----------
    smiles_to_graph: callable, str -> DGLGraph
        A function turning a SMILES string into a DGLGraph.
        Default to :func:`dgllife.utils.smiles_to_bigraph`.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    load : bool
        Whether to load the previously pre-processed dataset or pre-process from scratch.
        ``load`` should be False when we want to try different graph construction and
        featurization methods and need to preprocess from scratch. Default to True.
    log_every : bool
        Print a message every time ``log_every`` molecules are processed. Default to 1000.
    cache_file_path : str
        Path to the cached DGLGraphs, default to 'lipophilicity_dglgraph.bin'.

    Examples
    --------

    >>> from dgllife.data import Lipophilicity
    >>> from dgllife.utils import smiles_to_bigraph, CanonicalAtomFeaturizer

    >>> dataset = Lipophilicity(smiles_to_bigraph, CanonicalAtomFeaturizer())
    >>> # Get size of the dataset
    >>> len(dataset)
    4200
    >>> # Get the 0th datapoint, consisting of SMILES, DGLGraph and logD
    >>> dataset[0]
    ('Cn1c(CN2CCN(CC2)c3ccc(Cl)cc3)nc4ccccc14',
     DGLGraph(num_nodes=24, num_edges=54,
              ndata_schemes={'h': Scheme(shape=(74,), dtype=torch.float32)}
              edata_schemes={}),
     tensor([3.5400]))

    We also provide information for the ChEMBL id of the compound.

    >>> dataset.chembl_ids[i]

    We can also get the ChEMBL id along with SMILES, DGLGraph and logD at once.

    >>> dataset.load_full = True
    >>> dataset[0]
    ('Cn1c(CN2CCN(CC2)c3ccc(Cl)cc3)nc4ccccc14',
     DGLGraph(num_nodes=24, num_edges=54,
              ndata_schemes={'h': Scheme(shape=(74,), dtype=torch.float32)}
              edata_schemes={}),
     tensor([3.5400]),
     'CHEMBL596271')
    """
    def __init__(self,
                 smiles_to_graph=smiles_to_bigraph,
                 node_featurizer=None,
                 edge_featurizer=None,
                 load=True,
                 log_every=1000,
                 cache_file_path='lipophilicity_dglgraph.bin'):

        self._url = 'dataset/lipophilicity.zip'
        data_path = get_download_dir() + '/lipophilicity.zip'
        dir_path = get_download_dir() + '/lipophilicity'
        download(_get_dgl_url(self._url), path=data_path)
        extract_archive(data_path, dir_path)
        df = pd.read_csv(dir_path + '/Lipophilicity.csv')

        # ChEMBL ids
        self.chembl_ids = df['CMPD_CHEMBLID'].tolist()

        self.load_full = False

        super(Lipophilicity, self).__init__(df=df,
                                            smiles_to_graph=smiles_to_graph,
                                            node_featurizer=node_featurizer,
                                            edge_featurizer=edge_featurizer,
                                            smiles_column='smiles',
                                            cache_file_path=cache_file_path,
                                            task_names=['exp'],
                                            load=load,
                                            log_every=log_every,
                                            init_mask=False)

    def __getitem__(self, item):
        """Get datapoint with index

        Parameters
        ----------
        item : int
            Datapoint index

        Returns
        -------
        str
            SMILES for the ith datapoint
        DGLGraph
            DGLGraph for the ith datapoint
        Tensor of dtype float32 and shape (1)
            Labels of the ith datapoint
        str, optional
            ChEMBL id of the ith datapoint
        """
        if self.load_full:
            return self.smiles[item], self.graphs[item], self.labels[item], self.chembl_ids[item]
        else:
            return self.smiles[item], self.graphs[item], self.labels[item]



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Variant of Graph Convolutional Networks/Graph Isomorphism Networks in OGB's Examples

import dgl
import dgl.function as fn
import torch
import torch.nn as nn
import torch.nn.functional as F

from dgl.nn.pytorch.glob import SumPooling

__all__ = ['GNNOGB']

# pylint: disable=W0221, C0103
class GCNOGBLayer(nn.Module):
    r"""Variant of Single GCN layer from `Open Graph Benchmark: Datasets for Machine Learning on
    Graphs <https://arxiv.org/abs/2005.00687>`__

    Parameters
    ----------
    in_node_feats : int
        Number of input node features.
    in_edge_feats : int
        Number of input edge features.
    out_feats : int
        Number of output node features.
    """
    def __init__(self, in_node_feats, in_edge_feats, out_feats):
        super(GCNOGBLayer, self).__init__()

        self.project_in_node_feats = nn.Linear(in_node_feats, out_feats)
        self.project_in_edge_feats = nn.Linear(in_edge_feats, out_feats)
        self.project_residual = nn.Embedding(1, out_feats)

    def reset_parameters(self):
        """Reinitialize model parameters."""
        self.project_in_node_feats.reset_parameters()
        self.project_in_edge_feats.reset_parameters()
        self.project_residual.reset_parameters()

    def forward(self, g, node_feats, edge_feats, degs, norm):
        """Update node representations.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs
        node_feats : FloatTensor of shape (N, in_node_feats)
            Input node features. N for the total number of nodes in the batch of graphs.
        edge_feats : FloatTensor of shape (E, in_edge_feats)
            Input edge features. E for the total number of edges in the batch of graphs.
        degs : FloatTensor of shape (N, 1)
            Degree of each node in the batch of graphs.
            N for the total number of nodes in the batch of graphs.
        norm : FloatTensor of shape (E, 1)
            Edge-associated normalization coefficients. For an edge (i, j), this is
            equivalent to 1 / sqrt{d_i * d_j}, d_i and d_j are separately the degree
            of node i and j.

        Returns
        -------
        FloatTensor of shape (N, out_feats)
            Updated node representations.
        """
        g = g.local_var()
        node_feats = self.project_in_node_feats(node_feats)
        edge_feats = self.project_in_edge_feats(edge_feats)

        g.ndata['feat'] = node_feats
        g.apply_edges(fn.copy_u('feat', 'e'))
        edge_feats = F.relu(g.edata['e'] + edge_feats)
        g.edata['e'] = norm * edge_feats
        g.update_all(fn.copy_e('e', 'm'), fn.sum('m', 'feat'))

        residual_node_feats = node_feats + self.project_residual.weight
        residual_node_feats = F.relu(residual_node_feats)
        residual_node_feats = residual_node_feats * 1. / degs.view(-1, 1)

        rst = g.ndata['feat'] + residual_node_feats

        return rst

class GINOGBLayer(nn.Module):
    r"""Variant of Single GIN layer from `Open Graph Benchmark: Datasets for Machine Learning on
    Graphs <https://arxiv.org/abs/2005.00687>`__

    Parameters
    ----------
    node_feats : int
        Number of input and output node features.
    in_edge_feats : int
        Number of input edge features.
    """
    def __init__(self, node_feats, in_edge_feats):
        super(GINOGBLayer, self).__init__()

        self.project_in_edge_feats = nn.Linear(in_edge_feats, node_feats)
        self.eps = nn.Parameter(torch.Tensor([0]))
        self.project_out = nn.Sequential(
            nn.Linear(node_feats, 2 * node_feats),
            nn.BatchNorm1d(2 * node_feats),
            nn.ReLU(),
            nn.Linear(2 * node_feats, node_feats)
        )

    def reset_parameters(self):
        """Reinitialize model parameters."""
        self.project_in_edge_feats.reset_parameters()
        device = self.eps.device
        self.eps = nn.Parameter(torch.Tensor([0]).to(device))
        for layer in self.project_out:
            if isinstance(layer, (nn.Linear, nn.BatchNorm1d)):
                layer.reset_parameters()

    def forward(self, g, node_feats, edge_feats):
        """Update node representations.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs
        node_feats : FloatTensor of shape (N, node_feats)
            Input node features. N for the total number of nodes in the batch of graphs.
        edge_feats : FloatTensor of shape (E, in_edge_feats)
            Input edge features. E for the total number of edges in the batch of graphs.

        Returns
        -------
        FloatTensor of shape (N, node_feats)
            Updated node representations.
        """
        g = g.local_var()
        edge_feats = self.project_in_edge_feats(edge_feats)

        g.ndata['feat'] = node_feats
        g.apply_edges(fn.copy_u('feat', 'e'))
        g.edata['e'] = F.relu(edge_feats + g.edata['e'])
        g.update_all(fn.copy_e('e', 'm'), fn.sum('m', 'feat'))

        rst = g.ndata['feat']
        rst = self.project_out(rst + (1 + self.eps) * node_feats)

        return rst

class GNNOGB(nn.Module):
    r"""Variant of GCN/GIN from `Open Graph Benchmark: Datasets for Machine Learning on Graphs
    <https://arxiv.org/abs/2005.00687>`__

    Parameters
    ----------
    in_edge_feats : int
        Number of input edge features.
    num_node_types : int
        Number of node types to embed. (Default: 1)
    hidden_feats : int
        Size for hidden representations. (Default: 300)
    n_layers : int
        Number of GNN layers to use. (Default: 5)
    batchnorm : bool
        Whether to apply batch normalization. (Default: True)
    activation : callable or None
        Activation function to apply to the output of each GNN layer except for the last layer.
        If None, no activation will be applied. (Default: ReLU)
    dropout : float
        The probability for dropout. (Default: 0, i.e. no dropout)
    gnn_type : str
        The GNN type to use, which can be either 'gcn' or 'gin'. (Default: 'gcn')
    virtual_node : bool
        Whether to use virtual node. (Default: True)
    residual : bool
        Whether to apply residual connections for virtual node embeddings. (Default: False)
    jk : bool
        Whether to sum over the output of all GNN layers as in
        `JK networks <https://arxiv.org/abs/1806.03536>`__. (Default: False)
    """
    def __init__(self,
                 in_edge_feats,
                 num_node_types=1,
                 hidden_feats=300,
                 n_layers=5,
                 batchnorm=True,
                 activation=F.relu,
                 dropout=0.,
                 gnn_type='gcn',
                 virtual_node=True,
                 residual=False,
                 jk=False):
        super(GNNOGB, self).__init__()

        assert gnn_type in ['gcn', 'gin'], \
            "Expect gnn_type to be either 'gcn' or 'gin', got {}".format(gnn_type)

        self.n_layers = n_layers
        # Initial node embeddings
        self.node_encoder = nn.Embedding(num_node_types, hidden_feats)
        # Hidden layers
        self.layers = nn.ModuleList()
        self.gnn_type = gnn_type
        for _ in range(n_layers):
            if gnn_type == 'gcn':
                self.layers.append(GCNOGBLayer(in_node_feats=hidden_feats,
                                               in_edge_feats=in_edge_feats,
                                               out_feats=hidden_feats))
            else:
                self.layers.append(GINOGBLayer(node_feats=hidden_feats,
                                               in_edge_feats=in_edge_feats))

        self.virtual_node = virtual_node
        if virtual_node:
            self.virtual_node_emb = nn.Embedding(1, hidden_feats)
            self.mlp_virtual_project = nn.ModuleList()
            for _ in range(n_layers - 1):
                self.mlp_virtual_project.append(nn.Sequential(
                    nn.Linear(hidden_feats, 2 * hidden_feats),
                    nn.BatchNorm1d(2 * hidden_feats),
                    nn.ReLU(),
                    nn.Linear(2 * hidden_feats, hidden_feats),
                    nn.BatchNorm1d(hidden_feats),
                    nn.ReLU()))
            self.virtual_readout = SumPooling()

        if batchnorm:
            self.batchnorms = nn.ModuleList()
            for _ in range(n_layers):
                self.batchnorms.append(nn.BatchNorm1d(hidden_feats))
        else:
            self.batchnorms = None

        self.activation = activation
        self.dropout = nn.Dropout(dropout)
        self.residual = residual
        self.jk = jk

        self.reset_parameters()

    def reset_parameters(self):
        """Reinitialize model parameters."""
        self.node_encoder.reset_parameters()
        for gnn_layer in self.layers:
            gnn_layer.reset_parameters()

        if self.virtual_node:
            nn.init.constant_(self.virtual_node_emb.weight.data, 0)
            for mlp_layer in self.mlp_virtual_project:
                for layer in mlp_layer:
                    if isinstance(layer, (nn.Linear, nn.BatchNorm1d)):
                        layer.reset_parameters()

        if self.batchnorms is not None:
            for norm_layer in self.batchnorms:
                norm_layer.reset_parameters()

    def forward(self, g, node_feats, edge_feats):
        """Update node representations.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs
        node_feats : LongTensor of shape (N, 1)
            Input categorical node features. N for the number of nodes.
        edge_feats : FloatTensor of shape (E, in_edge_feats)
            Input edge features. E for the number of edges.

        Returns
        -------
        FloatTensor of shape (N, hidden_feats)
            Output node representations
        """
        if self.gnn_type == 'gcn':
            degs = (g.in_degrees().float() + 1).to(node_feats.device)
            norm = torch.pow(degs, -0.5).unsqueeze(-1)                # (N, 1)
            g.ndata['norm'] = norm
            g.apply_edges(fn.u_mul_v('norm', 'norm', 'norm'))
            norm = g.edata.pop('norm')

        if self.virtual_node:
            virtual_node_feats = self.virtual_node_emb(
                torch.zeros(g.batch_size).to(node_feats.dtype).to(node_feats.device))
        h_list = [self.node_encoder(node_feats)]

        for l in range(len(self.layers)):
            if self.virtual_node:
                virtual_feats_broadcast = dgl.broadcast_nodes(g, virtual_node_feats)
                h_list[l] = h_list[l] + virtual_feats_broadcast

            if self.gnn_type == 'gcn':
                h = self.layers[l](g, h_list[l], edge_feats, degs, norm)
            else:
                h = self.layers[l](g, h_list[l], edge_feats)

            if self.batchnorms is not None:
                h = self.batchnorms[l](h)

            if self.activation is not None and l != self.n_layers - 1:
                h = self.activation(h)
            h = self.dropout(h)
            h_list.append(h)

            if l < self.n_layers - 1 and self.virtual_node:
                ### Update virtual node representation from real node representations
                virtual_node_feats_tmp = self.virtual_readout(g, h_list[l]) + virtual_node_feats
                if self.residual:
                    virtual_node_feats = virtual_node_feats + self.dropout(
                        self.mlp_virtual_project[l](virtual_node_feats_tmp))
                else:
                    virtual_node_feats = self.dropout(
                        self.mlp_virtual_project[l](virtual_node_feats_tmp))

        if self.jk:
            return torch.stack(h_list, dim=0).sum(0)
        else:
            return h_list[-1]



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Weave
# pylint: disable= no-member, arguments-differ, invalid-name

import dgl.function as fn
import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ['WeaveGNN']

# pylint: disable=W0221, E1101
class WeaveLayer(nn.Module):
    r"""Single Weave layer from `Molecular Graph Convolutions: Moving Beyond Fingerprints
    <https://arxiv.org/abs/1603.00856>`__

    Parameters
    ----------
    node_in_feats : int
        Size for the input node features.
    edge_in_feats : int
        Size for the input edge features.
    node_node_hidden_feats : int
        Size for the hidden node representations in updating node representations.
        Default to 50.
    edge_node_hidden_feats : int
        Size for the hidden edge representations in updating node representations.
        Default to 50.
    node_out_feats : int
        Size for the output node representations. Default to 50.
    node_edge_hidden_feats : int
        Size for the hidden node representations in updating edge representations.
        Default to 50.
    edge_edge_hidden_feats : int
        Size for the hidden edge representations in updating edge representations.
        Default to 50.
    edge_out_feats : int
        Size for the output edge representations. Default to 50.
    activation : callable
        Activation function to apply. Default to ReLU.
    """
    def __init__(self,
                 node_in_feats,
                 edge_in_feats,
                 node_node_hidden_feats=50,
                 edge_node_hidden_feats=50,
                 node_out_feats=50,
                 node_edge_hidden_feats=50,
                 edge_edge_hidden_feats=50,
                 edge_out_feats=50,
                 activation=F.relu):
        super(WeaveLayer, self).__init__()

        self.activation = activation

        # Layers for updating node representations
        self.node_to_node = nn.Linear(node_in_feats, node_node_hidden_feats)
        self.edge_to_node = nn.Linear(edge_in_feats, edge_node_hidden_feats)
        self.update_node = nn.Linear(
            node_node_hidden_feats + edge_node_hidden_feats, node_out_feats)

        # Layers for updating edge representations
        self.left_node_to_edge = nn.Linear(node_in_feats, node_edge_hidden_feats)
        self.right_node_to_edge = nn.Linear(node_in_feats, node_edge_hidden_feats)
        self.edge_to_edge = nn.Linear(edge_in_feats, edge_edge_hidden_feats)
        self.update_edge = nn.Linear(
            2 * node_edge_hidden_feats + edge_edge_hidden_feats, edge_out_feats)

    def reset_parameters(self):
        """Reinitialize model parameters."""
        self.node_to_node.reset_parameters()
        self.edge_to_node.reset_parameters()
        self.update_node.reset_parameters()
        self.left_node_to_edge.reset_parameters()
        self.right_node_to_edge.reset_parameters()
        self.edge_to_edge.reset_parameters()
        self.update_edge.reset_parameters()

    def forward(self, g, node_feats, edge_feats, node_only=False):
        r"""Update node and edge representations.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs
        node_feats : float32 tensor of shape (V, node_in_feats)
            Input node features. V for the number of nodes in the batch of graphs.
        edge_feats : float32 tensor of shape (E, edge_in_feats)
            Input edge features. E for the number of edges in the batch of graphs.
        node_only : bool
            Whether to update node representations only. If False, edge representations
            will be updated as well. Default to False.

        Returns
        -------
        new_node_feats : float32 tensor of shape (V, node_out_feats)
            Updated node representations.
        new_edge_feats : float32 tensor of shape (E, edge_out_feats)
            Updated edge representations.
        """
        g = g.local_var()

        # Update node features
        node_node_feats = self.activation(self.node_to_node(node_feats))
        g.edata['e2n'] = self.activation(self.edge_to_node(edge_feats))
        g.update_all(fn.copy_edge('e2n', 'm'), fn.sum('m', 'e2n'))
        edge_node_feats = g.ndata.pop('e2n')
        new_node_feats = self.activation(self.update_node(
            torch.cat([node_node_feats, edge_node_feats], dim=1)))

        if node_only:
            return new_node_feats

        # Update edge features
        g.ndata['left_hv'] = self.left_node_to_edge(node_feats)
        g.ndata['right_hv'] = self.right_node_to_edge(node_feats)
        g.apply_edges(fn.u_add_v('left_hv', 'right_hv', 'first'))
        g.apply_edges(fn.u_add_v('right_hv', 'left_hv', 'second'))
        first_edge_feats = self.activation(g.edata.pop('first'))
        second_edge_feats = self.activation(g.edata.pop('second'))
        third_edge_feats = self.activation(self.edge_to_edge(edge_feats))
        new_edge_feats = self.activation(self.update_edge(
            torch.cat([first_edge_feats, second_edge_feats, third_edge_feats], dim=1)))

        return new_node_feats, new_edge_feats

class WeaveGNN(nn.Module):
    r"""The component of Weave for updating node and edge representations.

    Weave is introduced in `Molecular Graph Convolutions: Moving Beyond Fingerprints
    <https://arxiv.org/abs/1603.00856>`__.

    Parameters
    ----------
    node_in_feats : int
        Size for the input node features.
    edge_in_feats : int
        Size for the input edge features.
    num_layers : int
        Number of Weave layers to use, which is equivalent to the times of message passing.
        Default to 2.
    hidden_feats : int
        Size for the hidden node and edge representations. Default to 50.
    activation : callable
        Activation function to be used. It cannot be None. Default to ReLU.
    """
    def __init__(self,
                 node_in_feats,
                 edge_in_feats,
                 num_layers=2,
                 hidden_feats=50,
                 activation=F.relu):
        super(WeaveGNN, self).__init__()

        self.gnn_layers = nn.ModuleList()
        for i in range(num_layers):
            if i == 0:
                self.gnn_layers.append(WeaveLayer(node_in_feats=node_in_feats,
                                                  edge_in_feats=edge_in_feats,
                                                  node_node_hidden_feats=hidden_feats,
                                                  edge_node_hidden_feats=hidden_feats,
                                                  node_out_feats=hidden_feats,
                                                  node_edge_hidden_feats=hidden_feats,
                                                  edge_edge_hidden_feats=hidden_feats,
                                                  edge_out_feats=hidden_feats,
                                                  activation=activation))
            else:
                self.gnn_layers.append(WeaveLayer(node_in_feats=hidden_feats,
                                                  edge_in_feats=hidden_feats,
                                                  node_node_hidden_feats=hidden_feats,
                                                  edge_node_hidden_feats=hidden_feats,
                                                  node_out_feats=hidden_feats,
                                                  node_edge_hidden_feats=hidden_feats,
                                                  edge_edge_hidden_feats=hidden_feats,
                                                  edge_out_feats=hidden_feats,
                                                  activation=activation))

    def reset_parameters(self):
        """Reinitialize model parameters."""
        for layer in self.gnn_layers:
            layer.reset_parameters()

    def forward(self, g, node_feats, edge_feats, node_only=True):
        """Updates node representations (and edge representations).

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs.
        node_feats : float32 tensor of shape (V, node_in_feats)
            Input node features. V for the number of nodes in the batch of graphs.
        edge_feats : float32 tensor of shape (E, edge_in_feats)
            Input edge features. E for the number of edges in the batch of graphs.
        node_only : bool
            Whether to return updated node representations only or to return both
            node and edge representations. Default to True.

        Returns
        -------
        float32 tensor of shape (V, gnn_hidden_feats)
            Updated node representations.
        float32 tensor of shape (E, gnn_hidden_feats), optional
            This is returned only when ``node_only==False``. Updated edge representations.
        """
        for i in range(len(self.gnn_layers) - 1):
            node_feats, edge_feats = self.gnn_layers[i](g, node_feats, edge_feats)
        return self.gnn_layers[-1](g, node_feats, edge_feats, node_only)



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# GIN-based model for regression and classification on graphs.
# pylint: disable= no-member, arguments-differ, invalid-name

import torch.nn as nn

from dgl.nn.pytorch.glob import GlobalAttentionPooling, SumPooling, AvgPooling, MaxPooling

from ..gnn.gin import GIN

__all__ = ['GINPredictor']

# pylint: disable=W0221
class GINPredictor(nn.Module):
    """GIN-based model for regression and classification on graphs.

    GIN was first introduced in `How Powerful Are Graph Neural Networks
    <https://arxiv.org/abs/1810.00826>`__ for general graph property
    prediction problems. It was further extended in `Strategies for
    Pre-training Graph Neural Networks <https://arxiv.org/abs/1905.12265>`__
    for pre-training and semi-supervised learning on large-scale datasets.

    For classification tasks, the output will be logits, i.e. values before
    sigmoid or softmax.

    Parameters
    ----------
    num_node_emb_list : list of int
        num_node_emb_list[i] gives the number of items to embed for the
        i-th categorical node feature variables. E.g. num_node_emb_list[0] can be
        the number of atom types and num_node_emb_list[1] can be the number of
        atom chirality types.
    num_edge_emb_list : list of int
        num_edge_emb_list[i] gives the number of items to embed for the
        i-th categorical edge feature variables. E.g. num_edge_emb_list[0] can be
        the number of bond types and num_edge_emb_list[1] can be the number of
        bond direction types.
    num_layers : int
        Number of GIN layers to use. Default to 5.
    emb_dim : int
        The size of each embedding vector. Default to 300.
    JK : str
        JK for jumping knowledge as in `Representation Learning on Graphs with
        Jumping Knowledge Networks <https://arxiv.org/abs/1806.03536>`__. It decides
        how we are going to combine the all-layer node representations for the final output.
        There can be four options for this argument, ``'concat'``, ``'last'``, ``'max'`` and
        ``'sum'``. Default to 'last'.

        * ``'concat'``: concatenate the output node representations from all GIN layers
        * ``'last'``: use the node representations from the last GIN layer
        * ``'max'``: apply max pooling to the node representations across all GIN layers
        * ``'sum'``: sum the output node representations from all GIN layers
    dropout : float
        Dropout to apply to the output of each GIN layer. Default to 0.5.
    readout : str
        Readout for computing graph representations out of node representations, which
        can be ``'sum'``, ``'mean'``, ``'max'``, or ``'attention'``. Default to 'mean'.
    n_tasks : int
        Number of tasks, which is also the output size. Default to 1.
    """
    def __init__(self, num_node_emb_list, num_edge_emb_list, num_layers=5,
                 emb_dim=300, JK='last', dropout=0.5, readout='mean', n_tasks=1):
        super(GINPredictor, self).__init__()

        if num_layers < 2:
            raise ValueError('Number of GNN layers must be greater '
                             'than 1, got {:d}'.format(num_layers))

        self.gnn = GIN(num_node_emb_list=num_node_emb_list,
                       num_edge_emb_list=num_edge_emb_list,
                       num_layers=num_layers,
                       emb_dim=emb_dim,
                       JK=JK,
                       dropout=dropout)

        if readout == 'sum':
            self.readout = SumPooling()
        elif readout == 'mean':
            self.readout = AvgPooling()
        elif readout == 'max':
            self.readout = MaxPooling()
        elif readout == 'attention':
            if JK == 'concat':
                self.readout = GlobalAttentionPooling(
                    gate_nn=nn.Linear((num_layers + 1) * emb_dim, 1))
            else:
                self.readout = GlobalAttentionPooling(
                    gate_nn=nn.Linear(emb_dim, 1))
        else:
            raise ValueError("Expect readout to be 'sum', 'mean', "
                             "'max' or 'attention', got {}".format(readout))

        if JK == 'concat':
            self.predict = nn.Linear((num_layers + 1) * emb_dim, n_tasks)
        else:
            self.predict = nn.Linear(emb_dim, n_tasks)

    def forward(self, g, categorical_node_feats, categorical_edge_feats):
        """Graph-level regression/soft classification.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs
        categorical_node_feats : list of LongTensor of shape (N)
            * Input categorical node features
            * len(categorical_node_feats) should be the same as len(num_node_emb_list)
            * N is the total number of nodes in the batch of graphs
        categorical_edge_feats : list of LongTensor of shape (E)
            * Input categorical edge features
            * len(categorical_edge_feats) should be the same as
              len(num_edge_emb_list) in the arguments
            * E is the total number of edges in the batch of graphs

        Returns
        -------
        FloatTensor of shape (B, n_tasks)
            * Predictions on graphs
            * B for the number of graphs in the batch
        """
        node_feats = self.gnn(g, categorical_node_feats, categorical_edge_feats)
        graph_feats = self.readout(g, node_feats)
        return self.predict(graph_feats)



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# DGL-based package for applications in life science.

from .libinfo import __version__
from . import model

try:
    import rdkit
    from . import data
    from . import utils
except ImportError:
    print('RDKit is not installed, which is required for utils related to cheminformatics')



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .analysis import *
from .complex_to_graph import *
from .early_stop import *
from .eval import *
from .featurizers import *
from .mol_to_graph import *
from .io import *
from .splitters import *



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Readout for Weave
#
# pylint: disable= no-member, arguments-differ, invalid-name

import dgl
import torch
import torch.nn as nn

from torch.distributions import Normal

__all__ = ['WeaveGather']

# pylint: disable=W0221, E1101, E1102
class WeaveGather(nn.Module):
    r"""Readout in Weave

    Parameters
    ----------
    node_in_feats : int
        Size for the input node features.
    gaussian_expand : bool
        Whether to expand each dimension of node features by gaussian histogram.
        Default to True.
    gaussian_memberships : list of 2-tuples
        For each tuple, the first and second element separately specifies the mean
        and std for constructing a normal distribution. This argument comes into
        effect only when ``gaussian_expand==True``. By default, we set this to be
        ``[(-1.645, 0.283), (-1.080, 0.170), (-0.739, 0.134), (-0.468, 0.118),
        (-0.228, 0.114), (0., 0.114), (0.228, 0.114), (0.468, 0.118),
        (0.739, 0.134), (1.080, 0.170), (1.645, 0.283)]``.
    activation : callable
        Activation function to apply. Default to tanh.
    """
    def __init__(self,
                 node_in_feats,
                 gaussian_expand=True,
                 gaussian_memberships=None,
                 activation=nn.Tanh()):
        super(WeaveGather, self).__init__()

        self.gaussian_expand = gaussian_expand
        if gaussian_expand:
            if gaussian_memberships is None:
                gaussian_memberships = [
                    (-1.645, 0.283), (-1.080, 0.170), (-0.739, 0.134), (-0.468, 0.118),
                    (-0.228, 0.114), (0., 0.114), (0.228, 0.114), (0.468, 0.118),
                    (0.739, 0.134), (1.080, 0.170), (1.645, 0.283)]
            means, stds = map(list, zip(*gaussian_memberships))
            self.means = nn.ParameterList([
                nn.Parameter(torch.tensor(value), requires_grad=False)
                for value in means
            ])
            self.stds = nn.ParameterList([
                nn.Parameter(torch.tensor(value), requires_grad=False)
                for value in stds
            ])
            self.to_out = nn.Linear(node_in_feats * len(self.means), node_in_feats)
            self.activation = activation

    def gaussian_histogram(self, node_feats):
        r"""Constructs a gaussian histogram to capture the distribution of features

        Parameters
        ----------
        node_feats : float32 tensor of shape (V, node_in_feats)
            Input node features. V for the number of nodes in the batch of graphs.

        Returns
        -------
        float32 tensor of shape (V, node_in_feats * len(self.means))
            Updated node representations
        """
        gaussian_dists = [Normal(self.means[i], self.stds[i])
                          for i in range(len(self.means))]
        max_log_probs = [gaussian_dists[i].log_prob(self.means[i])
                         for i in range(len(self.means))]
        # Normalize the probabilities by the maximum point-wise probabilities,
        # whose results will be in range [0, 1]. Note that division of probabilities
        # is equivalent to subtraction of log probabilities and the latter one is cheaper.
        log_probs = [gaussian_dists[i].log_prob(node_feats) - max_log_probs[i]
                     for i in range(len(self.means))]
        probs = torch.stack(log_probs, dim=2).exp() # (V, node_in_feats, len(self.means))
        # Add a bias to avoid numerical issues in division
        probs = probs + 1e-7
        # Normalize the probabilities across all Gaussian distributions
        probs = probs / probs.sum(2, keepdim=True)

        return probs.reshape(node_feats.shape[0],
                             node_feats.shape[1] * len(self.means))

    def forward(self, g, node_feats):
        r"""Computes graph representations out of node representations.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs.
        node_feats : float32 tensor of shape (V, node_in_feats)
            Input node features. V for the number of nodes in the batch of graphs.

        Returns
        -------
        g_feats : float32 tensor of shape (G, node_in_feats)
            Output graph representations. G for the number of graphs in the batch.
        """
        if self.gaussian_expand:
            node_feats = self.gaussian_histogram(node_feats)

        with g.local_scope():
            g.ndata['h'] = node_feats
            g_feats = dgl.sum_nodes(g, 'h')

        if self.gaussian_expand:
            g_feats = self.to_out(g_feats)
            if self.activation is not None:
                g_feats = self.activation(g_feats)

        return g_feats



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# FreeSolv from MoleculeNet for the prediction of hydration free
# energy of small molecules in water

import pandas as pd

from dgl.data.utils import get_download_dir, download, _get_dgl_url, extract_archive

from .csv_dataset import MoleculeCSVDataset
from ..utils.mol_to_graph import smiles_to_bigraph

__all__ = ['FreeSolv']

class FreeSolv(MoleculeCSVDataset):
    r"""FreeSolv from MoleculeNet for the prediction of hydration free
    energy of small molecules in water

    The dataset provides experimental and calculated hydration free energy
    of small molecules in the water. The calculated values are derived from
    alchemical free energy calculations using molecular dynamics simulations.
    The experimental values are used for labels and the calculated values can
    be used for comparison.

    References:

        * [1] MoleculeNet: A Benchmark for Molecular Machine Learning.
        * [2] FreeSolv: a database of experimental and calculated hydration
              free energies, with input files.

    Parameters
    ----------
    smiles_to_graph: callable, str -> DGLGraph
        A function turning a SMILES string into a DGLGraph.
        Default to :func:`dgllife.utils.smiles_to_bigraph`.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    load : bool
        Whether to load the previously pre-processed dataset or pre-process from scratch.
        ``load`` should be False when we want to try different graph construction and
        featurization methods and need to preprocess from scratch. Default to True.
    log_every : bool
        Print a message every time ``log_every`` molecules are processed. Default to 1000.
    cache_file_path : str
        Path to the cached DGLGraphs, default to 'freesolv_dglgraph.bin'.

    Examples
    --------

    >>> from dgllife.data import FreeSolv
    >>> from dgllife.utils import smiles_to_bigraph, CanonicalAtomFeaturizer

    >>> dataset = FreeSolv(smiles_to_bigraph, CanonicalAtomFeaturizer())
    >>> # Get size of the dataset
    >>> len(dataset)
    642
    >>> # Get the 0th datapoint, consisting of SMILES, DGLGraph and hydration free energy
    >>> dataset[0]
    ('CN(C)C(=O)c1ccc(cc1)OC',
     DGLGraph(num_nodes=13, num_edges=26,
              ndata_schemes={'h': Scheme(shape=(74,), dtype=torch.float32)}
              edata_schemes={}),
     tensor([-11.0100]))

    We also provide information for the iupac name and calculated hydration free energy
    of the compound.

    >>> # Access the information mentioned above for the ith datapoint
    >>> dataset.iupac_names[i]
    >>> dataset.calc_energy[i]

    We can also get all these information along with SMILES, DGLGraph and hydration free
    energy at once.

    >>> dataset.load_full = True
    >>> dataset[0]
    ('CN(C)C(=O)c1ccc(cc1)OC',
     DGLGraph(num_nodes=13, num_edges=26,
              ndata_schemes={'h': Scheme(shape=(74,), dtype=torch.float32)}
              edata_schemes={}), tensor([-11.0100]),
     '4-methoxy-N,N-dimethyl-benzamide',
     -9.625)
    """
    def __init__(self,
                 smiles_to_graph=smiles_to_bigraph,
                 node_featurizer=None,
                 edge_featurizer=None,
                 load=True,
                 log_every=1000,
                 cache_file_path='freesolv_dglgraph.bin'):

        self._url = 'dataset/FreeSolv.zip'
        data_path = get_download_dir() + '/FreeSolv.zip'
        dir_path = get_download_dir() + '/FreeSolv'
        download(_get_dgl_url(self._url), path=data_path)
        extract_archive(data_path, dir_path)
        df = pd.read_csv(dir_path + '/SAMPL.csv')

        # Iupac names
        self.iupac_names = df['iupac'].tolist()
        # Calculated hydration free energy
        self.calc_energy = df['calc'].tolist()

        self.load_full = False

        super(FreeSolv, self).__init__(df=df,
                                       smiles_to_graph=smiles_to_graph,
                                       node_featurizer=node_featurizer,
                                       edge_featurizer=edge_featurizer,
                                       smiles_column='smiles',
                                       cache_file_path=cache_file_path,
                                       task_names=['expt'],
                                       load=load,
                                       log_every=log_every,
                                       init_mask=False)

    def __getitem__(self, item):
        """Get datapoint with index

        Parameters
        ----------
        item : int
            Datapoint index

        Returns
        -------
        str
            SMILES for the ith datapoint
        DGLGraph
            DGLGraph for the ith datapoint
        Tensor of dtype float32 and shape (1)
            Labels of the ith datapoint
        str, optional
            IUPAC nomenclature for the ith datapoint
        float, optional
            Calculated hydration free energy for the ith datapoint
        """
        if self.load_full:
            return self.smiles[item], self.graphs[item], self.labels[item], \
                   self.iupac_names[item], self.calc_energy[item]
        else:
            return self.smiles[item], self.graphs[item], self.labels[item]



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .alchemy import *
from .astrazeneca_chembl_solubility import *
from .csv_dataset import *
from .smiles_inference import *
from .pdbbind import *
from .pubchem_aromaticity import *
from .tox21 import *
from .esol import *
from .freesolv import *
from .lipophilicity import *
from .uspto import *



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Collection of model architectures

# Molecular property prediction
from .attentivefp_predictor import *
from .gat_predictor import *
from .gcn_predictor import *
from .gin_predictor import *
from .mgcn_predictor import *
from .mlp_predictor import *
from .mpnn_predictor import *
from .schnet_predictor import *
from .weave_predictor import *
from .gnn_ogb_predictor import *

# Generative models
try:
    # DGMG requires RDKit support
    from .dgmg import *
    from .jtnn import *
except ImportError:
    pass

# Reaction prediction
from .wln_reaction_center import *
from .wln_reaction_ranking import *

# Protein-Ligand Binding
from .acnn import *

# Link prediction
from .hadamard_link_predictor import *



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Graph neural networks for updating node representations

from .attentivefp import *
from .gat import *
from .gcn import *
from .gin import *
from .gnn_ogb import *
from .graphsage import *
from .mgcn import *
from .mpnn import *
from .schnet import *
from .weave import *
from .wln import *



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# USPTO for reaction prediction

import errno
import numpy as np
import os
import random
import torch

from collections import defaultdict
from copy import deepcopy
from dgl import DGLGraph
from dgl.data.utils import get_download_dir, download, _get_dgl_url, extract_archive, \
    save_graphs, load_graphs
from functools import partial
from itertools import combinations
from multiprocessing import Pool
from rdkit import Chem, RDLogger
from rdkit.Chem import rdmolops
from tqdm import tqdm

from ..utils.featurizers import BaseAtomFeaturizer, ConcatFeaturizer, one_hot_encoding, \
    atom_type_one_hot, atom_degree_one_hot, atom_explicit_valence_one_hot, \
    atom_implicit_valence_one_hot, atom_is_aromatic, atom_formal_charge_one_hot, \
    BaseBondFeaturizer, bond_type_one_hot, bond_is_conjugated, bond_is_in_ring
from ..utils.mol_to_graph import mol_to_bigraph, mol_to_complete_graph

__all__ = ['WLNCenterDataset',
           'USPTOCenter',
           'WLNRankDataset',
           'USPTORank']

# Disable RDKit warnings
RDLogger.DisableLog('rdApp.*')

# Atom types distinguished in featurization
atom_types = ['C', 'N', 'O', 'S', 'F', 'Si', 'P', 'Cl', 'Br', 'Mg', 'Na', 'Ca', 'Fe',
              'As', 'Al', 'I', 'B', 'V', 'K', 'Tl', 'Yb', 'Sb', 'Sn', 'Ag', 'Pd', 'Co',
              'Se', 'Ti', 'Zn', 'H', 'Li', 'Ge', 'Cu', 'Au', 'Ni', 'Cd', 'In', 'Mn', 'Zr',
              'Cr', 'Pt', 'Hg', 'Pb', 'W', 'Ru', 'Nb', 'Re', 'Te', 'Rh', 'Tc', 'Ba', 'Bi',
              'Hf', 'Mo', 'U', 'Sm', 'Os', 'Ir', 'Ce', 'Gd', 'Ga', 'Cs']

default_node_featurizer_center = BaseAtomFeaturizer({
    'hv': ConcatFeaturizer(
        [partial(atom_type_one_hot,
                 allowable_set=atom_types, encode_unknown=True),
         partial(atom_degree_one_hot,
                 allowable_set=list(range(5)), encode_unknown=True),
         partial(atom_explicit_valence_one_hot,
                 allowable_set=list(range(1, 6)), encode_unknown=True),
         partial(atom_implicit_valence_one_hot,
                 allowable_set=list(range(5)), encode_unknown=True),
         atom_is_aromatic]
    )
})

default_node_featurizer_rank = BaseAtomFeaturizer({
    'hv': ConcatFeaturizer(
        [partial(atom_type_one_hot,
                 allowable_set=atom_types, encode_unknown=True),
         partial(atom_formal_charge_one_hot,
                 allowable_set=[-3, -2, -1, 0, 1, 2], encode_unknown=True),
         partial(atom_degree_one_hot,
                 allowable_set=list(range(5)), encode_unknown=True),
         partial(atom_explicit_valence_one_hot,
                 allowable_set=list(range(1, 6)), encode_unknown=True),
         partial(atom_implicit_valence_one_hot,
                 allowable_set=list(range(5)), encode_unknown=True),
         atom_is_aromatic]
    )
})

default_edge_featurizer_center = BaseBondFeaturizer({
    'he': ConcatFeaturizer([
        bond_type_one_hot, bond_is_conjugated, bond_is_in_ring]
    )
})

default_edge_featurizer_rank = BaseBondFeaturizer({
    'he': ConcatFeaturizer([
        bond_type_one_hot, bond_is_in_ring]
    )
})

def default_atom_pair_featurizer(reactants):
    """Featurize each pair of atoms, which will be used in updating
    the edata of a complete DGLGraph.

    The features include the bond type between the atoms (if any) and whether
    they belong to the same molecule. It is used in the global attention mechanism.

    Parameters
    ----------
    reactants : str
        SMILES for reactants
    data_field : str
        Key for storing the features in DGLGraph.edata. Default to 'atom_pair'

    Returns
    -------
    float32 tensor of shape (V^2, 10)
        features for each pair of atoms.
    """
    # Decide the reactant membership for each atom
    atom_to_reactant = dict()
    reactant_list = reactants.split('.')
    for id, s in enumerate(reactant_list):
        mol = Chem.MolFromSmiles(s)
        for atom in mol.GetAtoms():
            atom_to_reactant[atom.GetIntProp('molAtomMapNumber') - 1] = id

    # Construct mapping from atom pair to RDKit bond object
    all_reactant_mol = Chem.MolFromSmiles(reactants)
    atom_pair_to_bond = dict()
    for bond in all_reactant_mol.GetBonds():
        atom1 = bond.GetBeginAtom().GetIntProp('molAtomMapNumber') - 1
        atom2 = bond.GetEndAtom().GetIntProp('molAtomMapNumber') - 1
        atom_pair_to_bond[(atom1, atom2)] = bond
        atom_pair_to_bond[(atom2, atom1)] = bond

    def _featurize_a_bond(bond):
        return bond_type_one_hot(bond) + bond_is_conjugated(bond) + bond_is_in_ring(bond)

    features = []
    num_atoms = all_reactant_mol.GetNumAtoms()
    for i in range(num_atoms):
        for j in range(num_atoms):
            pair_feature = np.zeros(10)
            if i == j:
                features.append(pair_feature)
                continue

            bond = atom_pair_to_bond.get((i, j), None)
            if bond is not None:
                pair_feature[1:7] = _featurize_a_bond(bond)
            else:
                pair_feature[0] = 1.
            pair_feature[-4] = 1. if atom_to_reactant[i] != atom_to_reactant[j] else 0.
            pair_feature[-3] = 1. if atom_to_reactant[i] == atom_to_reactant[j] else 0.
            pair_feature[-2] = 1. if len(reactant_list) == 1 else 0.
            pair_feature[-1] = 1. if len(reactant_list) > 1 else 0.
            features.append(pair_feature)

    return torch.from_numpy(np.stack(features, axis=0).astype(np.float32))

def get_pair_label(reactants_mol, graph_edits):
    """Construct labels for each pair of atoms in reaction center prediction

    Parameters
    ----------
    reactants_mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for all reactants in a reaction
    graph_edits : str
        Specifying which pairs of atoms loss a bond or form a particular bond in the reaction

    Returns
    -------
    float32 tensor of shape (V^2, 5)
        Labels constructed. V for the number of atoms in the reactants.
    """
    # 0 for losing the bond
    # 1, 2, 3, 1.5 separately for forming a single, double, triple or aromatic bond.
    bond_change_to_id = {0.0: 0, 1:1, 2:2, 3:3, 1.5:4}
    pair_to_changes = defaultdict(list)
    for edit in graph_edits.split(';'):
        a1, a2, change = edit.split('-')
        atom1 = int(a1) - 1
        atom2 = int(a2) - 1
        change = bond_change_to_id[float(change)]
        pair_to_changes[(atom1, atom2)].append(change)
        pair_to_changes[(atom2, atom1)].append(change)

    num_atoms = reactants_mol.GetNumAtoms()
    labels = torch.zeros((num_atoms, num_atoms, 5))
    for pair in pair_to_changes.keys():
        i, j = pair
        labels[i, j, pair_to_changes[(j, i)]] = 1.

    return labels.reshape(-1, 5)

def get_bond_changes(reaction):
    """Get the bond changes in a reaction.

    Parameters
    ----------
    reaction : str
        SMILES for a reaction, e.g. [CH3:14][NH2:15].[N+:1](=[O:2])([O-:3])[c:4]1[cH:5][c:6]([C:7]
        (=[O:8])[OH:9])[cH:10][cH:11][c:12]1[Cl:13].[OH2:16]>>[N+:1](=[O:2])([O-:3])[c:4]1[cH:5]
        [c:6]([C:7](=[O:8])[OH:9])[cH:10][cH:11][c:12]1[NH:15][CH3:14]. It consists of reactants,
        products and the atom mapping.

    Returns
    -------
    bond_changes : set of 3-tuples
        Each tuple consists of (atom1, atom2, change type)
        There are 5 possible values for change type. 0 for losing the bond, and 1, 2, 3, 1.5
        separately for forming a single, double, triple or aromatic bond.
    """
    reactants = Chem.MolFromSmiles(reaction.split('>')[0])
    products  = Chem.MolFromSmiles(reaction.split('>')[2])

    conserved_maps = [
        a.GetProp('molAtomMapNumber')
        for a in products.GetAtoms() if a.HasProp('molAtomMapNumber')]
    bond_changes = set() # keep track of bond changes

    # Look at changed bonds
    bonds_prev = {}
    for bond in reactants.GetBonds():
        nums = sorted(
            [bond.GetBeginAtom().GetProp('molAtomMapNumber'),
             bond.GetEndAtom().GetProp('molAtomMapNumber')])
        if (nums[0] not in conserved_maps) and (nums[1] not in conserved_maps):
            continue
        bonds_prev['{}~{}'.format(nums[0], nums[1])] = bond.GetBondTypeAsDouble()
    bonds_new = {}
    for bond in products.GetBonds():
        nums = sorted(
            [bond.GetBeginAtom().GetProp('molAtomMapNumber'),
             bond.GetEndAtom().GetProp('molAtomMapNumber')])
        bonds_new['{}~{}'.format(nums[0], nums[1])] = bond.GetBondTypeAsDouble()

    for bond in bonds_prev:
        if bond not in bonds_new:
            # lost bond
            bond_changes.add((bond.split('~')[0], bond.split('~')[1], 0.0))
        else:
            if bonds_prev[bond] != bonds_new[bond]:
                # changed bond
                bond_changes.add((bond.split('~')[0], bond.split('~')[1], bonds_new[bond]))
    for bond in bonds_new:
        if bond not in bonds_prev:
            # new bond
            bond_changes.add((bond.split('~')[0], bond.split('~')[1], bonds_new[bond]))

    return bond_changes

def process_line(line):
    """Process one line consisting of one reaction for working with WLN.

    Parameters
    ----------
    line : str
        One reaction in one line

    Returns
    -------
    formatted_reaction : str
        Formatted reaction
    """
    reaction = line.strip()
    bond_changes = get_bond_changes(reaction)
    formatted_reaction = '{} {}\n'.format(
        reaction, ';'.join(['{}-{}-{}'.format(x[0], x[1], x[2]) for x in bond_changes]))

    return formatted_reaction

def process_file(path, num_processes=1):
    """Pre-process a file of reactions for working with WLN.

    Parameters
    ----------
    path : str
        Path to the file of reactions
    num_processes : int
        Number of processes to use for data pre-processing. Default to 1.
    """
    with open(path, 'r') as input_file:
        lines = input_file.readlines()
    if num_processes == 1:
        results = []
        for li in lines:
            results.append(process_line(li))
    else:
        with Pool(processes=num_processes) as pool:
            results = pool.map(process_line, lines)
    with open(path + '.proc', 'w') as output_file:
        for line in results:
            output_file.write(line)
    print('Finished processing {}'.format(path))

def load_one_reaction(line):
    """Load one reaction and check if the reactants are valid.

    Parameters
    ----------
    line : str
        One reaction and the associated graph edits

    Returns
    -------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for the reactants. None will be returned if the
        reactants are not valid.
    reaction : str
        Reaction. None will be returned if the reactants are not valid.
    graph_edits : str
        Graph edits associated with the reaction. None will be returned if the
        reactants are not valid.
    """
    # Each line represents a reaction and the corresponding graph edits
    #
    # reaction example:
    # [CH3:14][OH:15].[NH2:12][NH2:13].[OH2:11].[n:1]1[n:2][cH:3][c:4]
    # ([C:7]([O:9][CH3:8])=[O:10])[cH:5][cH:6]1>>[n:1]1[n:2][cH:3][c:4]
    # ([C:7](=[O:9])[NH:12][NH2:13])[cH:5][cH:6]1
    # The reactants are on the left-hand-side of the reaction and the product
    # is on the right-hand-side of the reaction. The numbers represent atom mapping.
    #
    # graph_edits example:
    # 23-33-1.0;23-25-0.0
    # For a triplet a-b-c, a and b are the atoms that form or loss the bond.
    # c specifies the particular change, 0.0 for losing a bond, 1.0, 2.0, 3.0 and
    # 1.5 separately for forming a single, double, triple or aromatic bond.
    reaction, graph_edits = line.strip("\r\n ").split()
    reactants = reaction.split('>')[0]
    mol = Chem.MolFromSmiles(reactants)
    if mol is None:
        return None, None, None

    # Reorder atoms according to the order specified in the atom map
    atom_map_order = [-1 for _ in range(mol.GetNumAtoms())]
    for j in range(mol.GetNumAtoms()):
        atom = mol.GetAtomWithIdx(j)
        atom_map_order[atom.GetIntProp('molAtomMapNumber') - 1] = j
    mol = rdmolops.RenumberAtoms(mol, atom_map_order)

    return mol, reaction, graph_edits

def reaction_validity_full_check(reaction_file):
    """Tell valid reactions from invalid ones.

    Parameters
    ----------
    reaction_file : str
        Path to a file for reactions, where each line has the reaction SMILES and the
        corresponding graph edits.

    Returns
    -------
    valid_reactions : list
        Valid reactions for modeling
    invalid_reactions : list
        Invalid reactions for modeling
    """
    valid_reactions = []
    invalid_reactions = []
    with open(reaction_file, 'r') as file:
        for line in file:
            try:
                mol, reaction, graph_edits = load_one_reaction(line)
                assert mol is not None
                product_mol = Chem.MolFromSmiles(reaction.split('>')[2])
                assert product_mol is not None
                get_pair_label(mol, graph_edits)
                valid_reactions.append(line)
            except:
                invalid_reactions.append(line)

    return valid_reactions, invalid_reactions

class WLNCenterDataset(object):
    """Dataset for reaction center prediction with WLN

    Parameters
    ----------
    raw_file_path : str
        Path to the raw reaction file, where each line is the SMILES for a reaction.
        We will check if raw_file_path + '.proc' exists, where each line has the reaction
        SMILES and the corresponding graph edits. If not, we will preprocess
        the raw reaction file.
    mol_graph_path : str
        Path to save/load DGLGraphs for molecules.
    mol_to_graph: callable, str -> DGLGraph
        A function turning RDKit molecule instances into DGLGraphs.
        Default to :func:`dgllife.utils.mol_to_bigraph`.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. By default, we consider descriptors including atom type,
        atom degree, atom explicit valence, atom implicit valence, aromaticity.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. By default, we consider descriptors including bond type,
        whether bond is conjugated and whether bond is in ring.
    atom_pair_featurizer : callable, str -> dict
        Featurization for each pair of atoms in multiple reactants. The result will be
        used to update edata in the complete DGLGraphs. By default, the features include
        the bond type between the atoms (if any) and whether they belong to the same molecule.
    load : bool
        Whether to load the previously pre-processed dataset or pre-process from scratch.
        ``load`` should be False when we want to try different graph construction and
        featurization methods and need to preprocess from scratch. Default to True.
    num_processes : int
        Number of processes to use for data pre-processing. Default to 1.
    check_reaction_validity : bool
        Whether to check the validity of reactions before data pre-processing, which
        will introduce additional overhead. Default to True.
    reaction_validity_result_prefix : str or None
        Prefix for saving results for checking validity of reactions.
        This argument only comes into effect if ``check_reaction_validity`` is True,
        in which case we will save valid reactions in
        ``reaction_validity_result_prefix + _valid_reactions.proc`` and
        invalid ones in ``reaction_validity_result_prefix + _invalid_reactions.proc``.
        Default to ``''``.
    """
    def __init__(self,
                 raw_file_path,
                 mol_graph_path,
                 mol_to_graph=mol_to_bigraph,
                 node_featurizer=default_node_featurizer_center,
                 edge_featurizer=default_edge_featurizer_center,
                 atom_pair_featurizer=default_atom_pair_featurizer,
                 load=True,
                 num_processes=1,
                 check_reaction_validity=True,
                 reaction_validity_result_prefix=''):
        super(WLNCenterDataset, self).__init__()

        self._atom_pair_featurizer = atom_pair_featurizer
        self.atom_pair_features = []
        self.atom_pair_labels = []
        # Map number of nodes to a corresponding complete graph
        self.complete_graphs = dict()

        path_to_reaction_file = raw_file_path + '.proc'
        print('Pre-processing graph edits from reaction data')
        process_file(raw_file_path, num_processes)

        if check_reaction_validity:
            print('Start checking validity of input reactions for modeling...')
            valid_reactions, invalid_reactions = \
                reaction_validity_full_check(path_to_reaction_file)
            print('# valid reactions {:d}'.format(len(valid_reactions)))
            print('# invalid reactions {:d}'.format(len(invalid_reactions)))
            path_to_valid_reactions = reaction_validity_result_prefix + \
                                      '_valid_reactions.proc'
            path_to_invalid_reactions = reaction_validity_result_prefix + \
                                        '_invalid_reactions.proc'
            with open(path_to_valid_reactions, 'w') as f:
                for line in valid_reactions:
                    f.write(line)
            with open(path_to_invalid_reactions, 'w') as f:
                for line in invalid_reactions:
                    f.write(line)
            path_to_reaction_file = path_to_valid_reactions

        import time
        t0 = time.time()
        full_mols, full_reactions, full_graph_edits = \
            self.load_reaction_data(path_to_reaction_file, num_processes)
        print('Time spent', time.time() - t0)

        if load and os.path.isfile(mol_graph_path):
            print('Loading previously saved graphs...')
            self.reactant_mol_graphs, _ = load_graphs(mol_graph_path)
        else:
            print('Constructing graphs from scratch...')
            if num_processes == 1:
                self.reactant_mol_graphs = []
                for mol in full_mols:
                    self.reactant_mol_graphs.append(mol_to_graph(
                        mol, node_featurizer=node_featurizer,
                        edge_featurizer=edge_featurizer, canonical_atom_order=False))
            else:
                torch.multiprocessing.set_sharing_strategy('file_system')
                with Pool(processes=num_processes) as pool:
                    self.reactant_mol_graphs = pool.map(
                        partial(mol_to_graph, node_featurizer=node_featurizer,
                                edge_featurizer=edge_featurizer, canonical_atom_order=False),
                        full_mols)

            save_graphs(mol_graph_path, self.reactant_mol_graphs)

        self.mols = full_mols
        self.reactions = full_reactions
        self.graph_edits = full_graph_edits
        self.atom_pair_features.extend([None for _ in range(len(self.mols))])
        self.atom_pair_labels.extend([None for _ in range(len(self.mols))])

    def load_reaction_data(self, file_path, num_processes):
        """Load reaction data from the raw file.

        Parameters
        ----------
        file_path : str
            Path to read the file.
        num_processes : int
            Number of processes to use for data pre-processing.

        Returns
        -------
        all_mols : list of rdkit.Chem.rdchem.Mol
            RDKit molecule instances
        all_reactions : list of str
            Reactions
        all_graph_edits : list of str
            Graph edits in the reactions.
        """
        all_mols = []
        all_reactions = []
        all_graph_edits = []
        with open(file_path, 'r') as f:
            lines = f.readlines()
        if num_processes == 1:
            results = []
            for li in lines:
                mol, reaction, graph_edits = load_one_reaction(li)
                results.append((mol, reaction, graph_edits))
        else:
            with Pool(processes=num_processes) as pool:
                results = pool.map(load_one_reaction, lines)

        for mol, reaction, graph_edits in results:
            if mol is None:
                continue
            all_mols.append(mol)
            all_reactions.append(reaction)
            all_graph_edits.append(graph_edits)

        return all_mols, all_reactions, all_graph_edits

    def __len__(self):
        """Get the size for the dataset.

        Returns
        -------
        int
            Number of reactions in the dataset.
        """
        return len(self.mols)

    def __getitem__(self, item):
        """Get the i-th datapoint.

        Returns
        -------
        str
            Reaction
        str
            Graph edits for the reaction
        DGLGraph
            DGLGraph for the ith molecular graph
        DGLGraph
            Complete DGLGraph, which will be needed for predicting
            scores between each pair of atoms
        float32 tensor of shape (V^2, 10)
            Features for each pair of atoms.
        float32 tensor of shape (V^2, 5)
            Labels for reaction center prediction.
            V for the number of atoms in the reactants.
        """
        mol = self.mols[item]
        num_atoms = mol.GetNumAtoms()

        if num_atoms not in self.complete_graphs:
            self.complete_graphs[num_atoms] = mol_to_complete_graph(
                mol, add_self_loop=True, canonical_atom_order=False)

        if self.atom_pair_features[item] is None:
            reactants = self.reactions[item].split('>')[0]
            self.atom_pair_features[item] = self._atom_pair_featurizer(reactants)

        if self.atom_pair_labels[item] is None:
            self.atom_pair_labels[item] = get_pair_label(mol, self.graph_edits[item])

        return self.reactions[item], self.graph_edits[item], \
               self.reactant_mol_graphs[item], \
               self.complete_graphs[num_atoms], \
               self.atom_pair_features[item], \
               self.atom_pair_labels[item]

class USPTOCenter(WLNCenterDataset):
    """USPTO dataset for reaction center prediction.

    The dataset contains reactions from patents granted by United States Patent
    and Trademark Office (USPTO), collected by Lowe [1]. Jin et al. removes duplicates
    and erroneous reactions, obtaining a set of 480K reactions. They divide it
    into 400K, 40K, and 40K for training, validation and test.

    References:

        * [1] Patent reaction extraction
        * [2] Predicting Organic Reaction Outcomes with Weisfeiler-Lehman Network

    Parameters
    ----------
    subset : str
        Whether to use the training/validation/test set as in Jin et al.

        * 'train' for the training set
        * 'val' for the validation set
        * 'test' for the test set
    mol_to_graph: callable, str -> DGLGraph
        A function turning RDKit molecule instances into DGLGraphs.
        Default to :func:`dgllife.utils.mol_to_bigraph`.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. By default, we consider descriptors including atom type,
        atom degree, atom explicit valence, atom implicit valence, aromaticity.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. By default, we consider descriptors including bond type,
        whether bond is conjugated and whether bond is in ring.
    atom_pair_featurizer : callable, str -> dict
        Featurization for each pair of atoms in multiple reactants. The result will be
        used to update edata in the complete DGLGraphs. By default, the features include
        the bond type between the atoms (if any) and whether they belong to the same molecule.
    load : bool
        Whether to load the previously pre-processed dataset or pre-process from scratch.
        ``load`` should be False when we want to try different graph construction and
        featurization methods and need to preprocess from scratch. Default to True.
    num_processes : int
        Number of processes to use for data pre-processing. Default to 1.
    """
    def __init__(self,
                 subset,
                 mol_to_graph=mol_to_bigraph,
                 node_featurizer=default_node_featurizer_center,
                 edge_featurizer=default_edge_featurizer_center,
                 atom_pair_featurizer=default_atom_pair_featurizer,
                 load=True,
                 num_processes=1):
        assert subset in ['train', 'val', 'test'], \
            'Expect subset to be "train" or "val" or "test", got {}'.format(subset)
        print('Preparing {} subset of USPTO for reaction center prediction.'.format(subset))
        self._subset = subset
        if subset == 'val':
            subset = 'valid'

        self._url = 'dataset/uspto.zip'
        data_path = get_download_dir() + '/uspto.zip'
        extracted_data_path = get_download_dir() + '/uspto'
        download(_get_dgl_url(self._url), path=data_path)
        extract_archive(data_path, extracted_data_path)

        super(USPTOCenter, self).__init__(
            raw_file_path=extracted_data_path + '/{}.txt'.format(subset),
            mol_graph_path=extracted_data_path + '/{}_mol_graphs.bin'.format(subset),
            mol_to_graph=mol_to_graph,
            node_featurizer=node_featurizer,
            edge_featurizer=edge_featurizer,
            atom_pair_featurizer=atom_pair_featurizer,
            load=load,
            num_processes=num_processes,
            check_reaction_validity=False)

    @property
    def subset(self):
        """Get the subset used for USPTOCenter

        Returns
        -------
        str

            * 'full' for the complete dataset
            * 'train' for the training set
            * 'val' for the validation set
            * 'test' for the test set
        """
        return self._subset

def mkdir_p(path):
    """Create a folder for the given path.

    Parameters
    ----------
    path: str
        Folder to create
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

def load_one_reaction_rank(line):
    """Load one reaction and check if the reactants are valid.

    Parameters
    ----------
    line : str
        One reaction and the associated graph edits

    Returns
    -------
    reactants_mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for the reactants. None will be returned if the
        line is not valid.
    product_mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for the product. None will be returned if the line is not valid.
    reaction_real_bond_changes : list of 3-tuples
        Real bond changes in the reaction. Each tuple is of form (atom1, atom2, change_type). For
        change_type, 0.0 stands for losing a bond, 1.0, 2.0, 3.0 and 1.5 separately stands for
        forming a single, double, triple or aromatic bond.
    """
    # Each line represents a reaction and the corresponding graph edits
    #
    # reaction example:
    # [CH3:14][OH:15].[NH2:12][NH2:13].[OH2:11].[n:1]1[n:2][cH:3][c:4]
    # ([C:7]([O:9][CH3:8])=[O:10])[cH:5][cH:6]1>>[n:1]1[n:2][cH:3][c:4]
    # ([C:7](=[O:9])[NH:12][NH2:13])[cH:5][cH:6]1
    # The reactants are on the left-hand-side of the reaction and the product
    # is on the right-hand-side of the reaction. The numbers represent atom mapping.
    #
    # graph_edits example:
    # 23-33-1.0;23-25-0.0
    # For a triplet a-b-c, a and b are the atoms that form or loss the bond.
    # c specifies the particular change, 0.0 for losing a bond, 1.0, 2.0, 3.0 and
    # 1.5 separately for forming a single, double, triple or aromatic bond.
    reaction, graph_edits = line.strip("\r\n ").split()
    reactants, _, product = reaction.split('>')
    reactants_mol = Chem.MolFromSmiles(reactants)
    if reactants_mol is None:
        return None, None, None, None, None

    product_mol = Chem.MolFromSmiles(product)
    if product_mol is None:
        return None, None, None, None, None

    # Reorder atoms according to the order specified in the atom map
    atom_map_order = [-1 for _ in range(reactants_mol.GetNumAtoms())]
    for j in range(reactants_mol.GetNumAtoms()):
        atom = reactants_mol.GetAtomWithIdx(j)
        atom_map_order[atom.GetIntProp('molAtomMapNumber') - 1] = j
    reactants_mol = rdmolops.RenumberAtoms(reactants_mol, atom_map_order)

    reaction_real_bond_changes = []
    for changed_bond in graph_edits.split(';'):
        atom1, atom2, change_type = changed_bond.split('-')
        atom1, atom2 = int(atom1) - 1, int(atom2) - 1
        reaction_real_bond_changes.append(
            (min(atom1, atom2), max(atom1, atom2), float(change_type)))

    return reactants_mol, product_mol, reaction_real_bond_changes

def load_candidate_bond_changes_for_one_reaction(line):
    """Load candidate bond changes for a reaction

    Parameters
    ----------
    line : str
        Candidate bond changes separated by ;. Each candidate bond change takes the
        form of atom1, atom2, change_type and change_score.

    Returns
    -------
    list of 4-tuples
        Loaded candidate bond changes.
    """
    reaction_candidate_bond_changes = []
    elements = line.strip().split(';')[:-1]
    for candidate in elements:
        atom1, atom2, change_type, score = candidate.split(' ')
        atom1, atom2 = int(atom1) - 1, int(atom2) - 1
        reaction_candidate_bond_changes.append((
            min(atom1, atom2), max(atom1, atom2), float(change_type), float(score)))

    return reaction_candidate_bond_changes

def bookkeep_reactant(mol, candidate_pairs):
    """Bookkeep reaction-related information of reactants.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for reactants.
    candidate_pairs : list of 2-tuples
        Pairs of atoms that ranked high by a model for reaction center prediction.
        By assumption, the two atoms are different and the first atom has a smaller
        index than the second.

    Returns
    -------
    info : dict
        Reaction-related information of reactants
    """
    num_atoms = mol.GetNumAtoms()
    info = {
        # free valence of atoms
        'free_val': [0 for _ in range(num_atoms)],
        # Whether it is a carbon atom
        'is_c': [False for _ in range(num_atoms)],
        # Whether it is a carbon atom connected to a nitrogen atom in pyridine
        'is_c2_of_pyridine': [False for _ in range(num_atoms)],
        # Whether it is a phosphorous atom
        'is_p': [False for _ in range(num_atoms)],
        # Whether it is a sulfur atom
        'is_s': [False for _ in range(num_atoms)],
        # Whether it is an oxygen atom
        'is_o': [False for _ in range(num_atoms)],
        # Whether it is a nitrogen atom
        'is_n': [False for _ in range(num_atoms)],
        'pair_to_bond_val': dict(),
        'ring_bonds': set()
    }

    # bookkeep atoms
    for j, atom in enumerate(mol.GetAtoms()):
        info['free_val'][j] += atom.GetTotalNumHs() + abs(atom.GetFormalCharge())
        # An aromatic carbon atom next to an aromatic nitrogen atom can get a
        # carbonyl b/c of bookkeeping of hydroxypyridines
        if atom.GetSymbol() == 'C':
            info['is_c'][j] = True
            if atom.GetIsAromatic():
                for nbr in atom.GetNeighbors():
                    if nbr.GetSymbol() == 'N' and nbr.GetDegree() == 2:
                        info['is_c2_of_pyridine'][j] = True
                        break
        # A nitrogen atom should be allowed to become positively charged
        elif atom.GetSymbol() == 'N':
            info['free_val'][j] += 1 - atom.GetFormalCharge()
            info['is_n'][j] = True
        # Phosphorous atoms can form a phosphonium
        elif atom.GetSymbol() == 'P':
            info['free_val'][j] += 1 - atom.GetFormalCharge()
            info['is_p'][j] = True
        elif atom.GetSymbol() == 'O':
            info['is_o'][j] = True
        elif atom.GetSymbol() == 'S':
            info['is_s'][j] = True

    # bookkeep bonds
    for bond in mol.GetBonds():
        atom1, atom2 = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
        atom1, atom2 = min(atom1, atom2), max(atom1, atom2)
        type_val = bond.GetBondTypeAsDouble()
        info['pair_to_bond_val'][(atom1, atom2)] = type_val
        if (atom1, atom2) in candidate_pairs:
            info['free_val'][atom1] += type_val
            info['free_val'][atom2] += type_val
        if bond.IsInRing():
            info['ring_bonds'].add((atom1, atom2))

    return info

def bookkeep_product(mol):
    """Bookkeep reaction-related information of atoms/bonds in products

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for products.

    Returns
    -------
    info : dict
        Reaction-related information of atoms/bonds in products
    """
    info = {
        'atoms': set()
    }
    for atom in mol.GetAtoms():
        info['atoms'].add(atom.GetAtomMapNum() - 1)

    return info

def is_connected_change_combo(combo_ids, cand_change_adj):
    """Check whether the combo of bond changes yields a connected component.

    Parameters
    ----------
    combo_ids : tuple of int
        Ids for bond changes in the combination.
    cand_change_adj : bool ndarray of shape (N, N)
        Adjacency matrix for candidate bond changes. Two candidate bond
        changes are considered adjacent if they share a common atom.
        * N for the number of candidate bond changes.

    Returns
    -------
    bool
        Whether the combo of bond changes yields a connected component
    """
    if len(combo_ids) == 1:
        return True
    multi_hop_adj = np.linalg.matrix_power(
        cand_change_adj[combo_ids, :][:, combo_ids], len(combo_ids) - 1)
    # The combo is connected if the distance between
    # any pair of bond changes is within len(combo) - 1

    return np.all(multi_hop_adj)

def is_valid_combo(combo_changes, reactant_info):
    """Whether the combo of bond changes is chemically valid.

    Parameters
    ----------
    combo_changes : list of 4-tuples
        Each tuple consists of atom1, atom2, type of bond change (in the form of related
        valence) and score for the change.
    reactant_info : dict
        Reaction-related information of reactants

    Returns
    -------
    bool
        Whether the combo of bond changes is chemically valid.
    """
    num_atoms = len(reactant_info['free_val'])
    force_even_parity = np.zeros((num_atoms,), dtype=bool)
    force_odd_parity = np.zeros((num_atoms,), dtype=bool)
    pair_seen = defaultdict(bool)
    free_val_tmp = reactant_info['free_val'].copy()
    for (atom1, atom2, change_type, score) in combo_changes:
        if pair_seen[(atom1, atom2)]:
            # A pair of atoms cannot have two types of changes. Even if we
            # randomly pick one, that will be reduced to a combo of less changes
            return False
        pair_seen[(atom1, atom2)] = True

        # Special valence rules
        atom1_type_val = atom2_type_val = change_type
        if change_type == 2:
            # to form a double bond
            if reactant_info['is_o'][atom1]:
                if reactant_info['is_c2_of_pyridine'][atom2]:
                    atom2_type_val = 1.
                elif reactant_info['is_p'][atom2]:
                    # don't count information of =o toward valence
                    # but require odd valence parity
                    atom2_type_val = 0.
                    force_odd_parity[atom2] = True
                elif reactant_info['is_s'][atom2]:
                    atom2_type_val = 0.
                    force_even_parity[atom2] = True
            elif reactant_info['is_o'][atom2]:
                if reactant_info['is_c2_of_pyridine'][atom1]:
                    atom1_type_val = 1.
                elif reactant_info['is_p'][atom1]:
                    atom1_type_val = 0.
                    force_odd_parity[atom1] = True
                elif reactant_info['is_s'][atom1]:
                    atom1_type_val = 0.
                    force_even_parity[atom1] = True
            elif reactant_info['is_n'][atom1] and reactant_info['is_p'][atom2]:
                atom2_type_val = 0.
                force_odd_parity[atom2] = True
            elif reactant_info['is_n'][atom2] and reactant_info['is_p'][atom1]:
                atom1_type_val = 0.
                force_odd_parity[atom1] = True
            elif reactant_info['is_p'][atom1] and reactant_info['is_c'][atom2]:
                atom1_type_val = 0.
                force_odd_parity[atom1] = True
            elif reactant_info['is_p'][atom2] and reactant_info['is_c'][atom1]:
                atom2_type_val = 0.
                force_odd_parity[atom2] = True

        reactant_pair_val = reactant_info['pair_to_bond_val'].get((atom1, atom2), None)
        if reactant_pair_val is not None:
            free_val_tmp[atom1] += reactant_pair_val - atom1_type_val
            free_val_tmp[atom2] += reactant_pair_val - atom2_type_val
        else:
            free_val_tmp[atom1] -= atom1_type_val
            free_val_tmp[atom2] -= atom2_type_val

    free_val_tmp = np.array(free_val_tmp)
    # False if 1) too many connections 2) sulfur valence not even
    # 3) phosphorous valence not odd
    if any(free_val_tmp < 0) or \
            any(aval % 2 != 0 for aval in free_val_tmp[force_even_parity]) or \
            any(aval % 2 != 1 for aval in free_val_tmp[force_odd_parity]):
        return False
    return True

def edit_mol(reactant_mols, edits, product_info):
    """Simulate reaction via graph editing

    Parameters
    ----------
    reactant_mols : rdkit.Chem.rdchem.Mol
        RDKit molecule instances for reactants.
    edits : list of 4-tuples
        Bond changes for getting the product out of the reactants in a reaction.
        Each 4-tuple is of form (atom1, atom2, change_type, score), where atom1
        and atom2 are the end atoms to form or lose a bond, change_type is the
        type of bond change and score represents the confidence for the bond change
        by a model.
    product_info : dict
        proeduct_info['atoms'] gives a set of atom ids in the ground truth product molecule.

    Returns
    -------
    str
        SMILES for the main products
    """
    bond_change_to_type = {1: Chem.rdchem.BondType.SINGLE, 2: Chem.rdchem.BondType.DOUBLE,
                           3: Chem.rdchem.BondType.TRIPLE, 1.5: Chem.rdchem.BondType.AROMATIC}

    new_mol = Chem.RWMol(reactant_mols)
    [atom.SetNumExplicitHs(0) for atom in new_mol.GetAtoms()]

    for atom1, atom2, change_type, score in edits:
        bond = new_mol.GetBondBetweenAtoms(atom1, atom2)
        if bond is not None:
            new_mol.RemoveBond(atom1, atom2)
        if change_type > 0:
            new_mol.AddBond(atom1, atom2, bond_change_to_type[change_type])

    pred_mol = new_mol.GetMol()
    pred_smiles = Chem.MolToSmiles(pred_mol)
    pred_list = pred_smiles.split('.')
    pred_mols = []
    for pred_smiles in pred_list:
        mol = Chem.MolFromSmiles(pred_smiles)
        if mol is None:
            continue
        atom_set = set([atom.GetAtomMapNum() - 1 for atom in mol.GetAtoms()])
        if len(atom_set & product_info['atoms']) == 0:
            continue
        for atom in mol.GetAtoms():
            atom.SetAtomMapNum(0)
        pred_mols.append(mol)

    return '.'.join(sorted([Chem.MolToSmiles(mol) for mol in pred_mols]))

def get_product_smiles(reactant_mols, edits, product_info):
    """Get the product smiles of the reaction

    Parameters
    ----------
    reactant_mols : rdkit.Chem.rdchem.Mol
        RDKit molecule instances for reactants.
    edits : list of 4-tuples
        Bond changes for getting the product out of the reactants in a reaction.
        Each 4-tuple is of form (atom1, atom2, change_type, score), where atom1
        and atom2 are the end atoms to form or lose a bond, change_type is the
        type of bond change and score represents the confidence for the bond change
        by a model.
    product_info : dict
        proeduct_info['atoms'] gives a set of atom ids in the ground truth product molecule.

    Returns
    -------
    str
        SMILES for the main products
    """
    smiles = edit_mol(reactant_mols, edits, product_info)
    if len(smiles) != 0:
        return smiles
    try:
        Chem.Kekulize(reactant_mols)
    except Exception as e:
        return smiles
    return edit_mol(reactant_mols, edits, product_info)

def generate_valid_candidate_combos():
    return NotImplementedError

def pre_process_one_reaction(info, num_candidate_bond_changes, max_num_bond_changes,
                             max_num_change_combos, mode):
    """Pre-process one reaction for candidate ranking.

    Parameters
    ----------
    info : 4-tuple
        * candidate_bond_changes : list of tuples
            The candidate bond changes for the reaction
        * real_bond_changes : list of tuples
            The real bond changes for the reaction
        * reactant_mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance for reactants
        * product_mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance for product
    num_candidate_bond_changes : int
        Number of candidate bond changes to consider for the ground truth reaction.
    max_num_bond_changes : int
        Maximum number of bond changes per reaction.
    max_num_change_combos : int
        Number of bond change combos to consider for each reaction.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph.
    mode : str
        Whether the dataset is to be used for training, validation or test.

    Returns
    -------
    valid_candidate_combos : list
        valid_candidate_combos[i] gives a list of tuples, which is the i-th valid combo
        of candidate bond changes for the reaction.
    candidate_bond_changes : list of 4-tuples
        Refined candidate bond changes considered for combos.
    reactant_info : dict
        Reaction-related information of reactants.
    """
    assert mode in ['train', 'val', 'test'], \
        "Expect mode to be 'train' or 'val' or 'test', got {}".format(mode)
    candidate_bond_changes_, real_bond_changes, reactant_mol, product_mol = info
    candidate_pairs = [(atom1, atom2) for (atom1, atom2, _, _)
                       in candidate_bond_changes_]
    reactant_info = bookkeep_reactant(reactant_mol, candidate_pairs)
    if mode == 'train':
        product_info = bookkeep_product(product_mol)

    # Filter out candidate new bonds already in reactants
    candidate_bond_changes = []
    count = 0
    for (atom1, atom2, change_type, score) in candidate_bond_changes_:
        if ((atom1, atom2) not in reactant_info['pair_to_bond_val']) or \
                (reactant_info['pair_to_bond_val'][(atom1, atom2)] != change_type):
            candidate_bond_changes.append((atom1, atom2, change_type, score))
            count += 1
            if count == num_candidate_bond_changes:
                break

    # Check if two bond changes have atom in common
    cand_change_adj = np.eye(len(candidate_bond_changes), dtype=bool)
    for i in range(len(candidate_bond_changes)):
        atom1_1, atom1_2, _, _ = candidate_bond_changes[i]
        for j in range(i + 1, len(candidate_bond_changes)):
            atom2_1, atom2_2, _, _ = candidate_bond_changes[j]
            if atom1_1 == atom2_1 or atom1_1 == atom2_2 or \
                    atom1_2 == atom2_1 or atom1_2 == atom2_2:
                cand_change_adj[i, j] = cand_change_adj[j, i] = True

    # Enumerate combinations of k candidate bond changes and record
    # those that are connected and chemically valid
    valid_candidate_combos = []
    cand_change_ids = range(len(candidate_bond_changes))
    for k in range(1, max_num_bond_changes + 1):
        for combo_ids in combinations(cand_change_ids, k):
            # Check if the changed bonds form a connected component
            if not is_connected_change_combo(combo_ids, cand_change_adj):
                continue
            combo_changes = [candidate_bond_changes[j] for j in combo_ids]
            # Check if the combo is chemically valid
            if is_valid_combo(combo_changes, reactant_info):
                valid_candidate_combos.append(combo_changes)

    if mode == 'train':
        random.shuffle(valid_candidate_combos)
        # Index for the combo of candidate bond changes
        # that is equivalent to the gold combo
        real_combo_id = -1
        for j, combo_changes in enumerate(valid_candidate_combos):
            if set([(atom1, atom2, change_type) for
                    (atom1, atom2, change_type, score) in combo_changes]) == \
                    set(real_bond_changes):
                real_combo_id = j
                break

        # If we fail to find the real combo, make it the first entry
        if real_combo_id == -1:
            valid_candidate_combos = \
                [[(atom1, atom2, change_type, 0.0)
                  for (atom1, atom2, change_type) in real_bond_changes]] + \
                valid_candidate_combos
        else:
            valid_candidate_combos[0], valid_candidate_combos[real_combo_id] = \
                valid_candidate_combos[real_combo_id], valid_candidate_combos[0]

        product_smiles = get_product_smiles(
            reactant_mol, valid_candidate_combos[0], product_info)
        if len(product_smiles) > 0:
            # Remove combos yielding duplicate products
            product_smiles = set([product_smiles])
            new_candidate_combos = [valid_candidate_combos[0]]

            count = 0
            for combo in valid_candidate_combos[1:]:
                smiles = get_product_smiles(reactant_mol, combo, product_info)
                if smiles in product_smiles or len(smiles) == 0:
                    continue
                product_smiles.add(smiles)
                new_candidate_combos.append(combo)
                count += 1
                if count == max_num_change_combos:
                    break
            valid_candidate_combos = new_candidate_combos
    valid_candidate_combos = valid_candidate_combos[:max_num_change_combos]

    return valid_candidate_combos, candidate_bond_changes, reactant_info

def featurize_nodes_and_compute_combo_scores(
        node_featurizer, reactant_mol, valid_candidate_combos):
    """Featurize atoms in reactants and compute scores for combos of bond changes

    Parameters
    ----------
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph.
    reactant_mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for reactants in a reaction
    valid_candidate_combos : list
        valid_candidate_combos[i] gives a list of tuples, which is the i-th valid combo
        of candidate bond changes for the reaction.

    Returns
    -------
    node_feats : float32 tensor of shape (N, M)
        Node features for reactants, N for the number of nodes and M for the feature size
    combo_bias : float32 tensor of shape (B, 1)
        Scores for combos of bond changes, B equals len(valid_candidate_combos)
    """
    node_feats = node_featurizer(reactant_mol)['hv']
    combo_bias = torch.zeros(len(valid_candidate_combos), 1).float()
    for combo_id, combo in enumerate(valid_candidate_combos):
        combo_bias[combo_id] = sum([
            score for (atom1, atom2, change_type, score) in combo])

    return node_feats, combo_bias

def construct_graphs_rank(info, edge_featurizer):
    """Construct graphs for reactants and candidate products in a reaction and featurize
    their edges

    Parameters
    ----------
    info : 4-tuple
        * reactant_mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance for reactants in a reaction
        * candidate_combos : list
            candidate_combos[i] gives a list of tuples, which is the i-th valid combo
            of candidate bond changes for the reaction.
        * candidate_bond_changes : list of 4-tuples
            Refined candidate bond changes considered for candidate products
        * reactant_info : dict
            Reaction-related information of reactants.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph.

    Returns
    -------
    reaction_graphs : list of DGLGraphs
        DGLGraphs for reactants and candidate products with edge features in edata['he'],
        where the first graph is for reactants.
    """
    reactant_mol, candidate_combos, candidate_bond_changes, reactant_info = info
    # Graphs for reactants and candidate products
    reaction_graphs = []

    # Get graph for the reactants
    reactant_graph = mol_to_bigraph(reactant_mol, edge_featurizer=edge_featurizer,
                                    canonical_atom_order=False)
    reaction_graphs.append(reactant_graph)

    candidate_bond_changes_no_score = [
        (atom1, atom2, change_type)
        for (atom1, atom2, change_type, score) in candidate_bond_changes]

    # Prepare common components across all candidate products
    breaking_reactant_neighbors = []
    common_src_list = []
    common_dst_list = []
    common_edge_feats = []
    num_bonds = reactant_mol.GetNumBonds()
    for j in range(num_bonds):
        bond = reactant_mol.GetBondWithIdx(j)
        u = bond.GetBeginAtomIdx()
        v = bond.GetEndAtomIdx()
        u_sort, v_sort = min(u, v), max(u, v)
        # Whether a bond in reactants might get broken
        if (u_sort, v_sort, 0.0) not in candidate_bond_changes_no_score:
            common_src_list.extend([u, v])
            common_dst_list.extend([v, u])
            common_edge_feats.extend([reactant_graph.edata['he'][2 * j],
                                      reactant_graph.edata['he'][2 * j + 1]])
        else:
            breaking_reactant_neighbors.append((
                u_sort, v_sort, bond.GetBondTypeAsDouble()))

    for combo in candidate_combos:
        combo_src_list = deepcopy(common_src_list)
        combo_dst_list = deepcopy(common_dst_list)
        combo_edge_feats = deepcopy(common_edge_feats)
        candidate_bond_end_atoms = [
            (atom1, atom2) for (atom1, atom2, change_type, score) in combo]
        for (atom1, atom2, change_type) in breaking_reactant_neighbors:
            if (atom1, atom2) not in candidate_bond_end_atoms:
                # If a bond might be broken in some other combos but not this,
                # add it as a negative sample
                combo.append((atom1, atom2, change_type, 0.0))

        for (atom1, atom2, change_type, score) in combo:
            if change_type == 0:
                continue
            combo_src_list.extend([atom1, atom2])
            combo_dst_list.extend([atom2, atom1])
            feats = one_hot_encoding(change_type, [1.0, 2.0, 3.0, 1.5, -1])
            if (atom1, atom2) in reactant_info['ring_bonds']:
                feats[-1] = 1
            feats = torch.tensor(feats).float()
            combo_edge_feats.extend([feats, feats.clone()])

        combo_edge_feats = torch.stack(combo_edge_feats, dim=0)
        combo_graph = DGLGraph()
        combo_graph.add_nodes(reactant_graph.number_of_nodes())
        combo_graph.add_edges(combo_src_list, combo_dst_list)
        combo_graph.edata['he'] = combo_edge_feats
        reaction_graphs.append(combo_graph)

    return reaction_graphs

class WLNRankDataset(object):
    """Dataset for ranking candidate products with WLN

    Parameters
    ----------
    path_to_reaction_file : str
        Path to the processed reaction files, where each line has the reaction SMILES
        and the corresponding graph edits.
    candidate_bond_path : str
        Path to the candidate bond changes for product enumeration, where each line is
        candidate bond changes for a reaction by a WLN for reaction center prediction.
    mode : str
        'train', 'val', or 'test', indicating whether the dataset is used for training,
        validation or test.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. By default, we consider descriptors including atom type,
        atom formal charge, atom degree, atom explicit valence, atom implicit valence,
        aromaticity.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. By default, we consider descriptors including bond type
        and whether bond is in ring.
    size_cutoff : int
        By calling ``.ignore_large(True)``, we can optionally ignore reactions whose reactants
        contain more than ``size_cutoff`` atoms. Default to 100.
    max_num_changes_per_reaction : int
        Maximum number of bond changes per reaction. Default to 5.
    num_candidate_bond_changes : int
        Number of candidate bond changes to consider for each ground truth reaction.
        Default to 16.
    max_num_change_combos_per_reaction : int
        Number of bond change combos to consider for each reaction. Default to 150.
    num_processes : int
        Number of processes to use for data pre-processing. Default to 1.
    """
    def __init__(self,
                 path_to_reaction_file,
                 candidate_bond_path,
                 mode,
                 node_featurizer=default_node_featurizer_rank,
                 edge_featurizer=default_edge_featurizer_rank,
                 size_cutoff=100,
                 max_num_changes_per_reaction=5,
                 num_candidate_bond_changes=16,
                 max_num_change_combos_per_reaction=150,
                 num_processes=1):
        super(WLNRankDataset, self).__init__()

        assert mode in ['train', 'val', 'test'], \
            "Expect mode to be 'train' or 'val' or 'test', got {}".format(mode)
        self.mode = mode

        self.ignore_large_samples = False
        self.size_cutoff = size_cutoff

        self.reactant_mols, self.product_mols, self.real_bond_changes, \
        self.ids_for_small_samples = self.load_reaction_data(path_to_reaction_file, num_processes)
        self.candidate_bond_changes = self.load_candidate_bond_changes(candidate_bond_path)

        self.num_candidate_bond_changes = num_candidate_bond_changes
        self.max_num_changes_per_reaction = max_num_changes_per_reaction
        self.max_num_change_combos_per_reaction = max_num_change_combos_per_reaction
        self.node_featurizer = node_featurizer
        self.edge_featurizer = edge_featurizer

    def load_reaction_data(self, file_path, num_processes):
        """Load reaction data from the raw file.

        Parameters
        ----------
        file_path : str
            Path to read the file.
        num_processes : int
            Number of processes to use for data pre-processing.

        Returns
        -------
        all_reactant_mols : list of rdkit.Chem.rdchem.Mol
            RDKit molecule instances for reactants.
        all_product_mols : list of rdkit.Chem.rdchem.Mol
            RDKit molecule instances for products if the dataset is for training and
            None otherwise.
        all_real_bond_changes : list of list
            ``all_real_bond_changes[i]`` gives a list of tuples, which are ground
            truth bond changes for a reaction.
        ids_for_small_samples : list of int
            Indices for reactions whose reactants do not contain too many atoms
        """
        print('Stage 1/2: loading reaction data...')
        all_reactant_mols = []
        all_product_mols = []
        all_real_bond_changes = []
        ids_for_small_samples = []
        with open(file_path, 'r') as f:
            lines = f.readlines()

        def _update_from_line(id, loaded_result):
            reactants_mol, product_mol, reaction_real_bond_changes = loaded_result
            if reactants_mol is None:
                return
            all_product_mols.append(product_mol)
            all_reactant_mols.append(reactants_mol)
            all_real_bond_changes.append(reaction_real_bond_changes)
            if reactants_mol.GetNumAtoms() <= self.size_cutoff:
                ids_for_small_samples.append(id)

        if num_processes == 1:
            for id, li in enumerate(tqdm(lines)):
                loaded_line = load_one_reaction_rank(li)
                _update_from_line(id, loaded_line)
        else:
            with Pool(processes=num_processes) as pool:
                results = pool.map(
                    load_one_reaction_rank,
                    lines, chunksize=len(lines) // num_processes)
            for id in range(len(lines)):
                _update_from_line(id, results[id])

        return all_reactant_mols, all_product_mols, all_real_bond_changes, ids_for_small_samples

    def load_candidate_bond_changes(self, file_path):
        """Load candidate bond changes predicted by a WLN for reaction center prediction.

        Parameters
        ----------
        file_path : str
            Path to a file of candidate bond changes for each reaction.

        Returns
        -------
        all_candidate_bond_changes : list of list
            ``all_candidate_bond_changes[i]`` gives a list of tuples, which are candidate
            bond changes for a reaction.
        """
        print('Stage 2/2: loading candidate bond changes...')
        with open(file_path, 'r') as f:
            lines = f.readlines()

        all_candidate_bond_changes = []
        for li in tqdm(lines):
            all_candidate_bond_changes.append(
                load_candidate_bond_changes_for_one_reaction(li))

        return all_candidate_bond_changes

    def ignore_large(self, ignore=True):
        """Whether to ignore reactions where reactants contain too many atoms.

        Parameters
        ----------
        ignore : bool
            If ``ignore``, reactions where reactants contain too many atoms will be ignored.
        """
        self.ignore_large_samples = ignore

    def __len__(self):
        """Get the size for the dataset.

        Returns
        -------
        int
            Number of reactions in the dataset.
        """
        if self.ignore_large_samples:
            return len(self.ids_for_small_samples)
        else:
            return len(self.reactant_mols)

    def __getitem__(self, item):
        """Get the i-th datapoint.

        Parameters
        ----------
        item : int
            Index for the datapoint.

        Returns
        -------
        list of B + 1 DGLGraph
            The first entry in the list is the DGLGraph for the reactants and the rest are
            DGLGraphs for candidate products. Each DGLGraph has edge features in edata['he'] and
            node features in ndata['hv'].
        candidate_scores : float32 tensor of shape (B, 1)
            The sum of scores for bond changes in each combo, where B is the number of combos.
        labels : int64 tensor of shape (1, 1), optional
            Index for the true candidate product, which is always 0 with pre-processing. This is
            returned only when we are not in the training mode.
        valid_candidate_combos : list, optional
            valid_candidate_combos[i] gives a list of tuples, which is the i-th valid combo
            of candidate bond changes for the reaction. Each tuple is of form (atom1, atom2,
            change_type, score). atom1, atom2 are the atom mapping numbers - 1 of the two
            end atoms. change_type can be 0, 1, 2, 3, 1.5, separately for losing a bond, forming
            a single, double, triple, and aromatic bond.
        reactant_mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance for the reactants
        real_bond_changes : list of tuples
            Ground truth bond changes in a reaction. Each tuple is of form (atom1, atom2,
            change_type). atom1, atom2 are the atom mapping numbers - 1 of the two
            end atoms. change_type can be 0, 1, 2, 3, 1.5, separately for losing a bond, forming
            a single, double, triple, and aromatic bond.
        product_mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance for the product
        """
        if self.ignore_large_samples:
            item = self.ids_for_small_samples[item]

        raw_candidate_bond_changes = self.candidate_bond_changes[item]
        real_bond_changes = self.real_bond_changes[item]
        reactant_mol = self.reactant_mols[item]
        product_mol = self.product_mols[item]

        # Get valid candidate products, candidate bond changes considered and reactant info
        valid_candidate_combos, candidate_bond_changes, reactant_info = \
            pre_process_one_reaction(
                (raw_candidate_bond_changes, real_bond_changes,
                 reactant_mol, product_mol),
                self.num_candidate_bond_changes, self.max_num_changes_per_reaction,
                self.max_num_change_combos_per_reaction, self.mode)

        # Construct DGLGraphs and featurize their edges
        g_list = construct_graphs_rank(
            (reactant_mol, valid_candidate_combos,
             candidate_bond_changes, reactant_info),
            self.edge_featurizer)

        # Get node features and candidate scores
        node_feats, candidate_scores = featurize_nodes_and_compute_combo_scores(
            self.node_featurizer, reactant_mol, valid_candidate_combos)
        for g in g_list:
            g.ndata['hv'] = node_feats

        if self.mode == 'train':
            labels = torch.zeros(1, 1).long()
            return g_list, candidate_scores, labels
        else:
            reactant_mol = self.reactant_mols[item]
            real_bond_changes = self.real_bond_changes[item]
            product_mol = self.product_mols[item]
            return g_list, candidate_scores, valid_candidate_combos, \
                   reactant_mol, real_bond_changes, product_mol

class USPTORank(WLNRankDataset):
    """USPTO dataset for ranking candidate products.

    The dataset contains reactions from patents granted by United States Patent
    and Trademark Office (USPTO), collected by Lowe [1]. Jin et al. removes duplicates
    and erroneous reactions, obtaining a set of 480K reactions. They divide it
    into 400K, 40K, and 40K for training, validation and test.

    References:

        * [1] Patent reaction extraction
        * [2] Predicting Organic Reaction Outcomes with Weisfeiler-Lehman Network

    Parameters
    ----------
    subset : str
        Whether to use the training/validation/test set as in Jin et al.

        * 'train' for the training set
        * 'val' for the validation set
        * 'test' for the test set
    candidate_bond_path : str
        Path to the candidate bond changes for product enumeration, where each line is
        candidate bond changes for a reaction by a WLN for reaction center prediction.
    size_cutoff : int
        By calling ``.ignore_large(True)``, we can optionally ignore reactions whose reactants
        contain more than ``size_cutoff`` atoms. Default to 100.
    max_num_changes_per_reaction : int
        Maximum number of bond changes per reaction. Default to 5.
    num_candidate_bond_changes : int
        Number of candidate bond changes to consider for each ground truth reaction.
        Default to 16.
    max_num_change_combos_per_reaction : int
        Number of bond change combos to consider for each reaction. Default to 150.
    num_processes : int
        Number of processes to use for data pre-processing. Default to 1.
    """
    def __init__(self,
                 subset,
                 candidate_bond_path,
                 size_cutoff=100,
                 max_num_changes_per_reaction=5,
                 num_candidate_bond_changes=16,
                 max_num_change_combos_per_reaction=150,
                 num_processes=1):
        assert subset in ['train', 'val', 'test'], \
            'Expect subset to be "train" or "val" or "test", got {}'.format(subset)
        print('Preparing {} subset of USPTO for product candidate ranking.'.format(subset))
        self._subset = subset
        if subset == 'val':
            mode = 'val'
            subset = 'valid'
        else:
            mode = subset

        self._url = 'dataset/uspto.zip'
        data_path = get_download_dir() + '/uspto.zip'
        extracted_data_path = get_download_dir() + '/uspto'
        download(_get_dgl_url(self._url), path=data_path)
        extract_archive(data_path, extracted_data_path)

        super(USPTORank, self).__init__(
            path_to_reaction_file=extracted_data_path + '/{}.txt.proc'.format(subset),
            candidate_bond_path=candidate_bond_path,
            mode=mode,
            size_cutoff=size_cutoff,
            max_num_changes_per_reaction=max_num_changes_per_reaction,
            num_candidate_bond_changes=num_candidate_bond_changes,
            max_num_change_combos_per_reaction=max_num_change_combos_per_reaction,
            num_processes=num_processes)

    @property
    def subset(self):
        """Get the subset used for USPTOCenter

        Returns
        -------
        str

            * 'full' for the complete dataset
            * 'train' for the training set
            * 'val' for the validation set
            * 'test' for the test set
        """
        return self._subset



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Graph Isomorphism Networks.
# pylint: disable= no-member, arguments-differ, invalid-name

import dgl.function as fn
import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ['GIN']

# pylint: disable=W0221, C0103
class GINLayer(nn.Module):
    r"""Single Layer GIN from `Strategies for
    Pre-training Graph Neural Networks <https://arxiv.org/abs/1905.12265>`__

    Parameters
    ----------
    num_edge_emb_list : list of int
        num_edge_emb_list[i] gives the number of items to embed for the
        i-th categorical edge feature variables. E.g. num_edge_emb_list[0] can be
        the number of bond types and num_edge_emb_list[1] can be the number of
        bond direction types.
    emb_dim : int
        The size of each embedding vector.
    batch_norm : bool
        Whether to apply batch normalization to the output of message passing.
        Default to True.
    activation : None or callable
        Activation function to apply to the output node representations.
        Default to None.
    """
    def __init__(self, num_edge_emb_list, emb_dim, batch_norm=True, activation=None):
        super(GINLayer, self).__init__()

        self.mlp = nn.Sequential(
            nn.Linear(emb_dim, 2 * emb_dim),
            nn.ReLU(),
            nn.Linear(2 * emb_dim, emb_dim)
        )
        self.edge_embeddings = nn.ModuleList()
        for num_emb in num_edge_emb_list:
            emb_module = nn.Embedding(num_emb, emb_dim)
            self.edge_embeddings.append(emb_module)

        if batch_norm:
            self.bn = nn.BatchNorm1d(emb_dim)
        else:
            self.bn = None
        self.activation = activation
        self.reset_parameters()

    def reset_parameters(self):
        """Reinitialize model parameters."""
        for layer in self.mlp:
            if isinstance(layer, nn.Linear):
                layer.reset_parameters()

        for emb_module in self.edge_embeddings:
            nn.init.xavier_uniform_(emb_module.weight.data)

        if self.bn is not None:
            self.bn.reset_parameters()

    def forward(self, g, node_feats, categorical_edge_feats):
        """Update node representations.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs
        node_feats : FloatTensor of shape (N, emb_dim)
            * Input node features
            * N is the total number of nodes in the batch of graphs
            * emb_dim is the input node feature size, which must match emb_dim in initialization
        categorical_edge_feats : list of LongTensor of shape (E)
            * Input categorical edge features
            * len(categorical_edge_feats) should be the same as len(self.edge_embeddings)
            * E is the total number of edges in the batch of graphs

        Returns
        -------
        node_feats : float32 tensor of shape (N, emb_dim)
            Output node representations
        """
        edge_embeds = []
        for i, feats in enumerate(categorical_edge_feats):
            edge_embeds.append(self.edge_embeddings[i](feats))
        edge_embeds = torch.stack(edge_embeds, dim=0).sum(0)
        g = g.local_var()
        g.ndata['feat'] = node_feats
        g.edata['feat'] = edge_embeds
        g.update_all(fn.u_add_e('feat', 'feat', 'm'), fn.sum('m', 'feat'))

        node_feats = self.mlp(g.ndata.pop('feat'))
        if self.bn is not None:
            node_feats = self.bn(node_feats)
        if self.activation is not None:
            node_feats = self.activation(node_feats)

        return node_feats

class GIN(nn.Module):
    r"""Graph Isomorphism Network from `Strategies for
    Pre-training Graph Neural Networks <https://arxiv.org/abs/1905.12265>`__

    This module is for updating node representations only.

    Parameters
    ----------
    num_node_emb_list : list of int
        num_node_emb_list[i] gives the number of items to embed for the
        i-th categorical node feature variables. E.g. num_node_emb_list[0] can be
        the number of atom types and num_node_emb_list[1] can be the number of
        atom chirality types.
    num_edge_emb_list : list of int
        num_edge_emb_list[i] gives the number of items to embed for the
        i-th categorical edge feature variables. E.g. num_edge_emb_list[0] can be
        the number of bond types and num_edge_emb_list[1] can be the number of
        bond direction types.
    num_layers : int
        Number of GIN layers to use. Default to 5.
    emb_dim : int
        The size of each embedding vector. Default to 300.
    JK : str
        JK for jumping knowledge as in `Representation Learning on Graphs with
        Jumping Knowledge Networks <https://arxiv.org/abs/1806.03536>`__. It decides
        how we are going to combine the all-layer node representations for the final output.
        There can be four options for this argument, ``concat``, ``last``, ``max`` and ``sum``.
        Default to 'last'.

        * ``'concat'``: concatenate the output node representations from all GIN layers
        * ``'last'``: use the node representations from the last GIN layer
        * ``'max'``: apply max pooling to the node representations across all GIN layers
        * ``'sum'``: sum the output node representations from all GIN layers
    dropout : float
        Dropout to apply to the output of each GIN layer. Default to 0.5
    """
    def __init__(self, num_node_emb_list, num_edge_emb_list,
                 num_layers=5, emb_dim=300, JK='last', dropout=0.5):
        super(GIN, self).__init__()

        self.num_layers = num_layers
        self.JK = JK
        self.dropout = nn.Dropout(dropout)

        if num_layers < 2:
            raise ValueError('Number of GNN layers must be greater '
                             'than 1, got {:d}'.format(num_layers))

        self.node_embeddings = nn.ModuleList()
        for num_emb in num_node_emb_list:
            emb_module = nn.Embedding(num_emb, emb_dim)
            self.node_embeddings.append(emb_module)

        self.gnn_layers = nn.ModuleList()
        for layer in range(num_layers):
            if layer == num_layers - 1:
                self.gnn_layers.append(GINLayer(num_edge_emb_list, emb_dim))
            else:
                self.gnn_layers.append(GINLayer(num_edge_emb_list, emb_dim, activation=F.relu))

        self.reset_parameters()

    def reset_parameters(self):
        """Reinitialize model parameters."""
        for emb_module in self.node_embeddings:
            nn.init.xavier_uniform_(emb_module.weight.data)

        for layer in self.gnn_layers:
            layer.reset_parameters()

    def forward(self, g, categorical_node_feats, categorical_edge_feats):
        """Update node representations

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs
        categorical_node_feats : list of LongTensor of shape (N)
            * Input categorical node features
            * len(categorical_node_feats) should be the same as len(self.node_embeddings)
            * N is the total number of nodes in the batch of graphs
        categorical_edge_feats : list of LongTensor of shape (E)
            * Input categorical edge features
            * len(categorical_edge_feats) should be the same as
              len(num_edge_emb_list) in the arguments
            * E is the total number of edges in the batch of graphs

        Returns
        -------
        final_node_feats : float32 tensor of shape (N, M)
            Output node representations, N for the number of nodes and
            M for output size. In particular, M will be emb_dim * (num_layers + 1)
            if self.JK == 'concat' and emb_dim otherwise.
        """
        node_embeds = []
        for i, feats in enumerate(categorical_node_feats):
            node_embeds.append(self.node_embeddings[i](feats))
        node_embeds = torch.stack(node_embeds, dim=0).sum(0)

        all_layer_node_feats = [node_embeds]
        for layer in range(self.num_layers):
            node_feats = self.gnn_layers[layer](g, all_layer_node_feats[layer],
                                                categorical_edge_feats)
            node_feats = self.dropout(node_feats)
            all_layer_node_feats.append(node_feats)

        if self.JK == 'concat':
            final_node_feats = torch.cat(all_layer_node_feats, dim=1)
        elif self.JK == 'last':
            final_node_feats = all_layer_node_feats[-1]
        elif self.JK == 'max':
            all_layer_node_feats = [h.unsqueeze(0) for h in all_layer_node_feats]
            final_node_feats = torch.max(torch.cat(all_layer_node_feats, dim=0), dim=0)[0]
        elif self.JK == 'sum':
            all_layer_node_feats = [h.unsqueeze(0) for h in all_layer_node_feats]
            final_node_feats = torch.sum(torch.cat(all_layer_node_feats, dim=0), dim=0)
        else:
            return ValueError("Expect self.JK to be 'concat', 'last', "
                              "'max' or 'sum', got {}".format(self.JK))

        return final_node_feats



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Weave
# pylint: disable= no-member, arguments-differ, invalid-name

import torch.nn as nn
import torch.nn.functional as F

from ..gnn import WeaveGNN
from ..readout import WeaveGather

__all__ = ['WeavePredictor']

# pylint: disable=W0221
class WeavePredictor(nn.Module):
    r"""Weave for regression and classification on graphs.

    Weave is introduced in `Molecular Graph Convolutions: Moving Beyond Fingerprints
    <https://arxiv.org/abs/1603.00856>`__

    Parameters
    ----------
    node_in_feats : int
        Size for the input node features.
    edge_in_feats : int
        Size for the input edge features.
    num_gnn_layers : int
        Number of GNN (Weave) layers to use. Default to 2.
    gnn_hidden_feats : int
        Size for the hidden node and edge representations.
        Default to 50.
    gnn_activation : callable
        Activation function to be used in GNN (Weave) layers.
        Default to ReLU.
    graph_feats : int
        Size for the hidden graph representations. Default to 50.
    gaussian_expand : bool
        Whether to expand each dimension of node features by
        gaussian histogram in computing graph representations.
        Default to True.
    gaussian_memberships : list of 2-tuples
        For each tuple, the first and second element separately
        specifies the mean and std for constructing a normal
        distribution. This argument comes into effect only when
        ``gaussian_expand==True``. By default, we set this to be
        ``[(-1.645, 0.283), (-1.080, 0.170), (-0.739, 0.134),
        (-0.468, 0.118), (-0.228, 0.114), (0., 0.114),
        (0.228, 0.114), (0.468, 0.118), (0.739, 0.134),
        (1.080, 0.170), (1.645, 0.283)]``.
    readout_activation : callable
        Activation function to be used in computing graph
        representations out of node representations. Default to Tanh.
    n_tasks : int
        Number of tasks, which is also the output size. Default to 1.
    """
    def __init__(self,
                 node_in_feats,
                 edge_in_feats,
                 num_gnn_layers=2,
                 gnn_hidden_feats=50,
                 gnn_activation=F.relu,
                 graph_feats=128,
                 gaussian_expand=True,
                 gaussian_memberships=None,
                 readout_activation=nn.Tanh(),
                 n_tasks=1):
        super(WeavePredictor, self).__init__()

        self.gnn = WeaveGNN(node_in_feats=node_in_feats,
                            edge_in_feats=edge_in_feats,
                            num_layers=num_gnn_layers,
                            hidden_feats=gnn_hidden_feats,
                            activation=gnn_activation)
        self.node_to_graph = nn.Sequential(
            nn.Linear(gnn_hidden_feats, graph_feats),
            readout_activation,
            nn.BatchNorm1d(graph_feats)
        )
        self.readout = WeaveGather(node_in_feats=graph_feats,
                                   gaussian_expand=gaussian_expand,
                                   gaussian_memberships=gaussian_memberships,
                                   activation=readout_activation)
        self.predict = nn.Linear(graph_feats, n_tasks)

    def forward(self, g, node_feats, edge_feats):
        """Graph-level regression/soft classification.

        Parameters
        ----------
        g : DGLGraph
            DGLGraph for a batch of graphs.
        node_feats : float32 tensor of shape (V, node_in_feats)
            Input node features. V for the number of nodes.
        edge_feats : float32 tensor of shape (E, edge_in_feats)
            Input edge features. E for the number of edges.

        Returns
        -------
        float32 tensor of shape (G, n_tasks)
            Prediction for the graphs in the batch. G for the number of graphs.
        """
        node_feats = self.gnn(g, node_feats, edge_feats, node_only=True)
        node_feats = self.node_to_graph(node_feats)
        g_feats = self.readout(g, node_feats)

        return self.predict(g_feats)



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Information for the library.

# current version
__version__ = '0.2.4'



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Experimental solubility determined at AstraZeneca on a
# set of compounds, recorded in ChEMBL

import pandas as pd

from dgl.data.utils import get_download_dir, download, _get_dgl_url

from .csv_dataset import MoleculeCSVDataset
from ..utils.mol_to_graph import smiles_to_bigraph

__all__ = ['AstraZenecaChEMBLSolubility']

class AstraZenecaChEMBLSolubility(MoleculeCSVDataset):
    r"""Experimental solubility determined at AstraZeneca, extracted from ChEMBL

    The dataset provides experimental solubility (in nM unit) for 1763 molecules
    in pH7.4 using solid starting material using the method described in [1].

    References:

        * [1] A Highly Automated Assay for Determining the Aqueous Equilibrium
          Solubility of Drug Discovery Compounds
        * [2] `CHEMBL3301361 <https://www.ebi.ac.uk/chembl/document_report_card/CHEMBL3301361/>`__

    Parameters
    ----------
    smiles_to_graph: callable, str -> DGLGraph
        A function turning a SMILES string into a DGLGraph.
        Default to :func:`dgllife.utils.smiles_to_bigraph`.
    node_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for nodes like atoms in a molecule, which can be used to update
        ndata for a DGLGraph. Default to None.
    edge_featurizer : callable, rdkit.Chem.rdchem.Mol -> dict
        Featurization for edges like bonds in a molecule, which can be used to update
        edata for a DGLGraph. Default to None.
    load : bool
        Whether to load the previously pre-processed dataset or pre-process from scratch.
        ``load`` should be False when we want to try different graph construction and
        featurization methods and need to preprocess from scratch. Default to True.
    log_every : bool
        Print a message every time ``log_every`` molecules are processed. Default to 1000.
    cache_file_path : str
        Path to the cached DGLGraphs. Default to 'AstraZeneca_chembl_solubility_graph.bin'.
    log_of_values : bool
        Whether to take the logarithm of the solubility values. Before taking the logarithm,
        the values can have a range of [100, 1513600]. After taking the logarithm, the
        values will have a range of [4.61, 14.23]. Default to True.

    Examples
    --------

    >>> from dgllife.data import AstraZenecaChEMBLSolubility
    >>> from dgllife.utils import smiles_to_bigraph, CanonicalAtomFeaturizer

    >>> dataset = AstraZenecaChEMBLSolubility(smiles_to_bigraph, CanonicalAtomFeaturizer())
    >>> # Get size of the dataset
    >>> len(dataset)
    1763
    >>> # Get the 0th datapoint, consisting of SMILES, DGLGraph and solubility
    >>> dataset[0]
    ('Cc1nc(C)c(-c2ccc([C@H]3CC[C@H](Cc4nnn[nH]4)CC3)cc2)nc1C(N)=O',
     DGLGraph(num_nodes=29, num_edges=64,
              ndata_schemes={'h': Scheme(shape=(74,), dtype=torch.float32)}
              edata_schemes={}),
     tensor([12.5032]))

    We also provide information for the ChEMBL id and molecular weight of the compound.

    >>> dataset.chembl_ids[i]
    >>> dataset.mol_weight[i]

    We can also get the ChEMBL id and molecular weight along with SMILES, DGLGraph and
    solubility at once.

    >>> dataset.load_full = True
    >>> dataset[0]
    ('Cc1nc(C)c(-c2ccc([C@H]3CC[C@H](Cc4nnn[nH]4)CC3)cc2)nc1C(N)=O',
     DGLGraph(num_nodes=29, num_edges=64,
              ndata_schemes={'h': Scheme(shape=(74,), dtype=torch.float32)}
              edata_schemes={}),
     tensor([12.5032]),
     'CHEMBL2178940',
     391.48)
    """
    def __init__(self,
                 smiles_to_graph=smiles_to_bigraph,
                 node_featurizer=None,
                 edge_featurizer=None,
                 load=True,
                 log_every=1000,
                 cache_file_path='AstraZeneca_chembl_solubility_graph.bin',
                 log_of_values=True):

        self._url = 'dataset/AstraZeneca_ChEMBL_Solubility.csv'
        data_path = get_download_dir() + '/AstraZeneca_ChEMBL_Solubility.csv'
        download(_get_dgl_url(self._url), path=data_path)
        df = pd.read_csv(data_path)

        # ChEMBL ids
        self.chembl_ids = df['Molecule ChEMBL ID'].tolist()
        # Molecular weight
        self.mol_weight = df['Molecular Weight'].tolist()

        self.load_full = False

        super(AstraZenecaChEMBLSolubility, self).__init__(
            df=df,
            smiles_to_graph=smiles_to_graph,
            node_featurizer=node_featurizer,
            edge_featurizer=edge_featurizer,
            smiles_column='Smiles',
            cache_file_path=cache_file_path,
            task_names=['Solubility'],
            load=load,
            log_every=log_every,
            init_mask=False)

        if log_of_values:
            self.labels = self.labels.log()

    def __getitem__(self, item):
        """Get datapoint with index

        Parameters
        ----------
        item : int
            Datapoint index

        Returns
        -------
        str
            SMILES for the ith datapoint
        DGLGraph
            DGLGraph for the ith datapoint
        Tensor of dtype float32 and shape (1)
            Labels of the ith datapoint
        str, optional
            ChEMBL id of the ith datapoint
        float, optional
            Molecular weight of the ith datapoint
        """
        if self.load_full:
            return self.smiles[item], self.graphs[item], self.labels[item], \
                   self.chembl_ids[item], self.mol_weight[item]
        else:
            return self.smiles[item], self.graphs[item], self.labels[item]



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Readout functions for computing molecular representations
# out of node and edge representations.

from .attentivefp_readout import *
from .mlp_readout import *
from .weave_readout import *
from .weighted_sum_and_max import *



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Predictor for link prediction by taking elementwise multiplication of node representations

import torch.nn as nn
import torch.nn.functional as F

__all__ = ['HadamardLinkPredictor']

class HadamardLinkPredictor(nn.Module):
    """Link prediction by taking the elementwise multiplication of two node representations

    The elementwise multiplication is also called Hadamard product.

    Parameters
    ----------
    in_feats : int
        Number of input node features
    hidden_feats : int
        Number of hidden features. Default to 256.
    num_layers : int
        Number of linear layers used in total, which should be
        at least 2, counting the input and output layers. Default to 3.
    n_tasks : int
        Number of output tasks. Default to 1.
    dropout : float
        Dropout before each linear layer except for the first one.
        Default to 0., i.e. no dropout is performed.
    activation : callable
        Activation function to apply after the output of each linear layer.
        Default to ReLU.
    """
    def __init__(self,
                 in_feats,
                 hidden_feats=256,
                 num_layers=3,
                 n_tasks=1,
                 dropout=0.,
                 activation=F.relu):
        super(HadamardLinkPredictor, self).__init__()

        assert num_layers >= 2, 'Expect num_layers to be at least 2, got {:d}'.format(num_layers)

        self.layers = nn.ModuleList()
        # input layer
        self.layers.append(nn.Linear(in_feats, hidden_feats))
        # hidden layers
        for _ in range(num_layers - 2):
            self.layers.append(nn.Linear(hidden_feats, hidden_feats))
        # output layer
        self.layers.append(nn.Linear(hidden_feats, n_tasks))
        self.dropout = nn.Dropout(dropout)
        self.activation = activation

    def reset_parameters(self):
        # Reset the parameters of the Linear layers
        for layer in self.layers:
            layer.reset_parameters()

    def forward(self, left_node_feats, right_node_feats):
        """Link Prediction

        Perform link prediction for P pairs of nodes. Note
        that this model is symmetric and we don't have
        separate parameters for the two arguments.

        Parameters
        ----------
        left_node_feats : float32 tensor of shape (P, D1)
            Representations for the first node in P pairs.
            D1 for the number of input node features.
        right_node_feats : float32 tensor of shape (P, D1)
            Representations for the second node in P pairs.
            D1 for the number of input node features.

        Returns
        -------
        float32 tensor of shape (P, D2)
            Pre-softmax/sigmoid logits, D2 equals n_tasks.
        """
        pair_feats = left_node_feats * right_node_feats
        for layer in self.layers[:-1]:
            pair_feats = layer(pair_feats)
            if self.activation is not None:
                pair_feats = self.activation(pair_feats)
            pair_feats = self.dropout(pair_feats)
        out = self.layers[-1](pair_feats)

        return out
